/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.jmx;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.ObjectName;

import org.apache.commons.io.IOUtils;
import org.ow2.petals.admin.api.ArtifactAdministration;
import org.ow2.petals.admin.api.artifact.Artifact;
import org.ow2.petals.admin.api.artifact.ArtifactState.State;
import org.ow2.petals.admin.api.artifact.Component;
import org.ow2.petals.admin.api.artifact.Component.ComponentType;
import org.ow2.petals.admin.api.artifact.ServiceAssembly;
import org.ow2.petals.admin.api.artifact.ServiceUnit;
import org.ow2.petals.admin.api.artifact.SharedLibrary;
import org.ow2.petals.admin.api.artifact.lifecycle.ComponentLifecycle;
import org.ow2.petals.admin.api.artifact.lifecycle.ServiceAssemblyLifecycle;
import org.ow2.petals.admin.api.conf.ArtifactUrlRewriter;
import org.ow2.petals.admin.api.exception.ArtifactAdministrationException;
import org.ow2.petals.admin.api.exception.ArtifactNotFoundException;
import org.ow2.petals.admin.api.exception.ArtifactTypeIsNeededException;
import org.ow2.petals.admin.api.exception.ArtifactVersionIsNeededException;
import org.ow2.petals.admin.api.exception.UnsupportedArtifactTypeException;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jmx.api.api.AdminServiceClient;
import org.ow2.petals.jmx.api.api.DeploymentServiceClient;
import org.ow2.petals.jmx.api.api.InstallationServiceClient;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.exception.AdminDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.AdminServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.DeploymentServiceDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.DeploymentServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.InstallationServiceDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.InstallationServiceErrorException;

/**
 * 
 * @author Nicolas Oddoux - EBM WebSourcing
 */
public class JMXArtifactAdministration extends ArtifactAdministration {

    private final JMXPetalsAdministration admin;

    protected JMXArtifactAdministration(final JMXPetalsAdministration admin,
            final ArtifactUrlRewriter artifactUrlRewriter) {
        super(admin, artifactUrlRewriter);
        this.admin = admin;
    }

    @Override
    public void stopAndUndeployAllArtifacts() throws ArtifactAdministrationException {
        try {
            JMXClient client = admin.getJMXClient();
            DeploymentServiceClient deploymentServiceClient = client.getDeploymentServiceClient();
            deploymentServiceClient.stopAllServiceAssemblies();
            deploymentServiceClient.shutdownAllServiceAssemblies();
            deploymentServiceClient.undeployAllServiceAssemblies(true);
            InstallationServiceClient installationServiceClient = client
                    .getInstallationServiceClient();
            installationServiceClient.stopAllComponents();
            installationServiceClient.shutdownAllComponents();
            installationServiceClient.uninstallAllComponents();
            installationServiceClient.unloadAllInstallers(true);
            installationServiceClient.uninstallAllSharedLibrary();
        } catch (ConnectionErrorException cee) {
            throw new ArtifactAdministrationException(cee);
        } catch (DeploymentServiceErrorException dsee) {
            throw new ArtifactAdministrationException(dsee);
        } catch (DeploymentServiceDoesNotExistException dsdnee) {
            throw new ArtifactAdministrationException(dsdnee);
        } catch (InstallationServiceDoesNotExistException isdnee) {
            throw new ArtifactAdministrationException(isdnee);
        } catch (InstallationServiceErrorException isee) {
            throw new ArtifactAdministrationException(isee);
        }
    }

    @Override
    public List<Artifact> listArtifacts() throws ArtifactAdministrationException {
        try {
            final List<Artifact> artifacts = new ArrayList<>();

            final JMXClient client = admin.getJMXClient();

            final Map<String, Component> components = new HashMap<>();

            // Retrieve components BC and SE
            final ObjectName[] bindingComponents = client.getAdminServiceClient().getBindingComponents();
            for (final ObjectName bindingComponent : bindingComponents) {
                final String name = bindingComponent.getKeyProperty("name");
                final Component component = new Component(name, ComponentType.BC);
                final ComponentLifecycle componentLifecycle = lifecycleFactory.createComponentLifecycle(component);
                componentLifecycle.updateState();
                components.put(name, component);
            }

            final ObjectName[] serviceEngines = client.getAdminServiceClient().getEngineComponents();
            for (final ObjectName serviceEngine : serviceEngines) {
                final String name = serviceEngine.getKeyProperty("name");
                final Component component = new Component(name, ComponentType.SE);
                final ComponentLifecycle componentLifecycle = lifecycleFactory.createComponentLifecycle(component);
                componentLifecycle.updateState();
                components.put(name, component);
            }

            // Retrieve components that have only their installer loaded
            final String[] installers = client.getInstallationServiceClient().getInstallers();
            for (final String installer : installers) {
                // We check the name first because it is more differential
                // than the type of artifact
                if (!components.containsKey(installer)) {
                    final Component component;
                    if (client.getAdminServiceClient().isBinding(installer)) {
                        component = new Component(installer, ComponentType.BC, State.LOADED);
                    } else if (client.getAdminServiceClient().isEngine(installer)) {
                        component = new Component(installer, ComponentType.SE, State.LOADED);
                    } else {
                        // This exception should not occurs
                        throw new ArtifactAdministrationException(
                                "Unable to determine the type of the component for installer '"
                                        + installer + "'.");
                    }
                    components.put(installer, component);
                }
            }

            artifacts.addAll(components.values());

            // Retrieve shared libraries
            final String[][] sls = client.getInstallationServiceClient().getInstalledSharedLibraries();
            for (final String[] sl : sls) {
                String name = sl[0];
                String version = sl[1];
                artifacts.add(new SharedLibrary(name, version));
                for (String componentName : client.getInstallationServiceClient()
                        .getInstalledComponentsForSharedlibrary(name, version)) {
                    Component component = components.get(componentName);
                    if (component != null) {
                        // we should not use the same object!
                        component.getSharedLibraries().add(new SharedLibrary(name, version));
                    } else {
                        throw new ArtifactAdministrationException(
                                "Unknown component " + componentName + " from SL " + name);
                    }
                }
            }

            // Retrieve service assemblies
            final String[] saNames = client.getDeploymentServiceClient().getDeployedServiceAssemblies();
            for (final String saName : saNames) {
                final ServiceAssembly serviceAssembly = createServiceAssembly(client, saName);
                final ServiceAssemblyLifecycle saLifecycle = lifecycleFactory
                        .createServiceAssemblyLifecycle(serviceAssembly);
                saLifecycle.updateState();
                artifacts.add(serviceAssembly);
            }

            return artifacts;
        } catch (final ConnectionErrorException cee) {
            throw new ArtifactAdministrationException(cee);
        } catch (final AdminServiceErrorException asee) {
            throw new ArtifactAdministrationException(asee);
        } catch (final AdminDoesNotExistException adnee) {
            throw new ArtifactAdministrationException(adnee);
        } catch (final DeploymentServiceErrorException dsee) {
            throw new ArtifactAdministrationException(dsee);
        } catch (final DeploymentServiceDoesNotExistException dsdnee) {
            throw new ArtifactAdministrationException(dsdnee);
        } catch (final InstallationServiceErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (final InstallationServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    @Override
    public Artifact getArtifact(final String type, final String name, final String version)
            throws UnsupportedArtifactTypeException, ArtifactNotFoundException, ArtifactTypeIsNeededException,
            ArtifactVersionIsNeededException, ArtifactAdministrationException {

        final Artifact artifact;
        if (type != null) {
            if (ComponentType.BC.toString().equals(type)) {
                artifact = this.getArtifactComponent(name, ComponentType.BC);
            } else if (ComponentType.SE.toString().equals(type)) {
                artifact = this.getArtifactComponent(name, ComponentType.SE);
            } else if (SharedLibrary.TYPE.equals(type)) {
                artifact = this.getArtifactSL(name, version);
            } else if (ServiceAssembly.TYPE.equals(type)) {
                artifact = this.getArtifactSA(name);
            } else {
                throw new UnsupportedArtifactTypeException(type);
            }

            if (artifact == null) {
                throw new ArtifactNotFoundException(type, name, version);
            }
        }
        else {
            final Artifact artifactBC = this.getArtifactComponent(name, ComponentType.BC);
            final Artifact artifactSE = this.getArtifactComponent(name, ComponentType.SE);
            final Artifact artifactSL = this.getArtifactSL(name, version);
            final Artifact artifactSA = this.getArtifactSA(name);
            if (artifactBC != null && artifactSE == null && artifactSL == null && artifactSA == null) {
                artifact = artifactBC;
            }
            else if (artifactBC == null && artifactSE != null && artifactSL == null && artifactSA == null) {
                artifact = artifactSE;
            }
            else if (artifactBC == null && artifactSE == null && artifactSL != null && artifactSA == null) {
                artifact = artifactSL;
            }
            else if (artifactBC == null && artifactSE == null && artifactSL == null && artifactSA != null) {
                artifact = artifactSA;
            }
            else if (artifactBC == null && artifactSE == null && artifactSL == null && artifactSA == null) {
                throw new ArtifactNotFoundException(type, name, version);
            }
            else {
                // More than one artifact match the name. 
                throw new ArtifactTypeIsNeededException(name);
            }
        }

        return artifact;
    }

    private Artifact getArtifactComponent(final String name, ComponentType type)
            throws ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            AdminServiceClient admin = client.getAdminServiceClient();
            final ObjectName[] components = type == ComponentType.BC ? admin.getBindingComponents()
                    : admin.getEngineComponents();

            boolean found = false;
            for (final ObjectName component : components) {
                if (name.equals(component.getKeyProperty("name"))) {
                    found = true;
                    break;
                }
            }

            // Here, no component with the given name is installed. We check if a
            // component installer was only loaded.
            if (!found && (type == ComponentType.BC ? admin.isBinding(name) : admin.isEngine(name))) {
                // A component installer exists for this binding component
                found = true;
            }

            if (found) {
                Component component = new Component(name, type);
                retrieveSharedLibraries(client, component);
                return component;
            } else {
                return null;
            }
        } catch (final ConnectionErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (final AdminServiceErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (final AdminDoesNotExistException | InstallationServiceErrorException
                | InstallationServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    private void retrieveSharedLibraries(JMXClient client, Component component)
            throws InstallationServiceErrorException, InstallationServiceDoesNotExistException,
            ConnectionErrorException {
        final String[][] sls = client.getInstallationServiceClient().getInstalledSharedLibraries();
        for (final String[] sl : sls) {
            String name = sl[0];
            String version = sl[1];
            for (String componentName : client.getInstallationServiceClient()
                    .getInstalledComponentsForSharedlibrary(name, version)) {
                if (componentName.equals(component.getName())) {
                    component.getSharedLibraries().add(new SharedLibrary(name, version));
                }
            }
        }
    }

    private Artifact getArtifactSL(final String askedSlName, final String askedSlVersion)
            throws ArtifactAdministrationException {

        try {
            final JMXClient client = admin.getJMXClient();
            final String[][] sls = client.getInstallationServiceClient().getInstalledSharedLibraries();
            if (askedSlVersion == null) {
                boolean slNameFound = false;
                String slVersion = null;
                for (final String[] sl : sls) {
                    if (sl[0].equals(askedSlName)) {
                        if (!slNameFound) {
                            slNameFound = true;
                            slVersion = sl[1];
                        } else {
                            // More than one shared library matches the given name (several versions exist)
                            throw new ArtifactVersionIsNeededException(askedSlName);
                        }
                    }
                }
                if (slNameFound) {
                    return new SharedLibrary(askedSlName, slVersion);
                } else {
                    return null;
                }
            } else {
                SharedLibrary result = null;
                for (final String[] sl : sls) {
                    if (sl[0].equals(askedSlName) && sl[1].equals(askedSlVersion)) {
                        result = new SharedLibrary(askedSlName, askedSlVersion);
                        break;
                    }
                }
                return result;
            }
        } catch (final ConnectionErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (final InstallationServiceErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (final InstallationServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    private Artifact getArtifactSA(final String name) throws ArtifactAdministrationException {

        try {
            final JMXClient client = admin.getJMXClient();
            final String[] saNames = client.getDeploymentServiceClient()
                    .getDeployedServiceAssemblies();
            for (final String saName : saNames) {
                if (saName.equals(name)) {
                    return createServiceAssembly(client, saName);
                }
            }
            return null;
        } catch (final ConnectionErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (final DeploymentServiceErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (final DeploymentServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    private ServiceAssembly createServiceAssembly(JMXClient client, String saName)
            throws ArtifactAdministrationException {
        try {
            String descriptor;
            List<ServiceUnit> serviceUnits;
            descriptor = client.getDeploymentServiceClient().getServiceAssemblyDescriptor(saName);

            final Jbi saJbiDescriptor;
            ByteArrayInputStream bais = new ByteArrayInputStream(descriptor.getBytes());
            try {
                saJbiDescriptor = JBIDescriptorBuilder.getInstance().buildJavaJBIDescriptor(bais);
            } finally {
                IOUtils.closeQuietly(bais);
            }

            serviceUnits = new ArrayList<>();
            List<org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit> suDescriptors = saJbiDescriptor
                    .getServiceAssembly().getServiceUnit();
            for (org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit suDescriptor : suDescriptors) {
                String suName = suDescriptor.getIdentification().getName();
                String suTargetComponentName = suDescriptor.getTarget().getComponentName();
                ServiceUnit serviceUnit = new ServiceUnit(suName, suTargetComponentName);
                serviceUnits.add(serviceUnit);
            }
            return new ServiceAssembly(saName, serviceUnits);
        } catch (Exception e) {
            throw new ArtifactAdministrationException(e);
        }

    }
}
