/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.jmx;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.ow2.petals.admin.api.ContainerAdministration;
import org.ow2.petals.admin.api.PetalsAdministration;
import org.ow2.petals.admin.api.artifact.Artifact;
import org.ow2.petals.admin.api.artifact.Component;
import org.ow2.petals.admin.api.artifact.Logger;
import org.ow2.petals.admin.api.artifact.ServiceAssembly;
import org.ow2.petals.admin.api.artifact.SharedLibrary;
import org.ow2.petals.admin.api.exception.ArtifactAdministrationException;
import org.ow2.petals.admin.api.exception.ContainerAdministrationException;
import org.ow2.petals.admin.api.exception.InvalidLogLevelException;
import org.ow2.petals.admin.api.exception.LoggerNotFoundException;
import org.ow2.petals.admin.api.exception.NoConnectionException;
import org.ow2.petals.admin.jmx.topology.TopologyUtils;
import org.ow2.petals.admin.topology.Container;
import org.ow2.petals.admin.topology.Container.PortType;
import org.ow2.petals.admin.topology.Domain;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.clientserverapi.system.logging.exception.LoggingServiceException;
import org.ow2.petals.clientserverapi.topology.RemoteContainer;
import org.ow2.petals.clientserverapi.topology.exception.TopologyServiceException;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.exception.AdminDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.AdminServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.LoggerServiceDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.LoggerServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.NotRemoteJMXClientException;
import org.ow2.petals.jmx.api.api.exception.PetalsAdminDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.PetalsAdminServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.TopologyServiceDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.TopologyServiceErrorException;

/**
 * 
 * @author Nicolas Oddoux - EBM WebSourcing
 */
public class JMXContainerAdministration implements ContainerAdministration {

    private final JMXPetalsAdministration admin;

    protected JMXContainerAdministration(final JMXPetalsAdministration admin) {
        this.admin = admin;
    }

    @Override
    public String getSystemInfo() throws ContainerAdministrationException {
        try {
            JMXClient client = admin.getJMXClient();
            return client.getAdminServiceClient().getSystemInfo();
        } catch (ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (AdminServiceErrorException asee) {
            throw new ContainerAdministrationException(asee);
        } catch (AdminDoesNotExistException adnee) {
            throw new ContainerAdministrationException(adnee);
        }
    }

    @Override
    public void stopContainer() throws ContainerAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            client.getPetalsAdminServiceClient().stopContainer();
            admin.disconnect();
        } catch (final ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (final PetalsAdminServiceErrorException pasee) {
            throw new ContainerAdministrationException(pasee);
        } catch (final PetalsAdminDoesNotExistException padnee) {
            throw new ContainerAdministrationException(padnee);
        } catch (final PetalsException e) {
            throw new ContainerAdministrationException(e);
        }
    }

    @Override
    public void shutdownContainer() throws ContainerAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            client.getPetalsAdminServiceClient().shutdownContainer();
            admin.disconnect();
        } catch (final ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (final PetalsAdminServiceErrorException pasee) {
            throw new ContainerAdministrationException(pasee);
        } catch (final PetalsAdminDoesNotExistException padnee) {
            throw new ContainerAdministrationException(padnee);
        } catch (final PetalsException e) {
            throw new ContainerAdministrationException(e);
        }
    }

    @Override
    public Domain getTopology(final String containerFilterRegexp, final String securityPassPhrase,
            final boolean retrieveJbiArtifacts)
            throws ContainerAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            // Note: retrieveJbiArtifacts implies sensibleInformationRequired because we need to connect to
            // every container
            final Set<Map<String, String>> topology = client.getTopologyServiceClient().retrieveTopology(
                    securityPassPhrase, retrieveJbiArtifacts);
            final Domain domain = TopologyUtils.createListOfDomains(topology, containerFilterRegexp);
            if (retrieveJbiArtifacts) {
                for (final Container container : domain.getContainers()) {
                    this.retrieveJbiArtifacts(container);
                }
            }
            return domain;
        } catch (final ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (final TopologyServiceErrorException pasee) {
            throw new ContainerAdministrationException(pasee);
        } catch (final TopologyServiceDoesNotExistException padnee) {
            throw new ContainerAdministrationException(padnee);
        } catch (final PetalsException e) {
            throw new ContainerAdministrationException(e);
        }
    }

    @Override
    public Domain getTopology(final String securityPassPhrase, final boolean retrieveJbiArtifacts)
            throws ContainerAdministrationException {
        return getTopology(null, securityPassPhrase, retrieveJbiArtifacts);
    }

    private void retrieveJbiArtifacts(final Container container) throws ContainerAdministrationException {
        try {
            final boolean isAlreadyConnected;
            final JMXClient initialJmxClient = admin.getJMXClient();
            if (InetAddress.getByName(container.getHost()).equals(initialJmxClient.getHost())
                    && container.getPorts().get(PortType.JMX).intValue() == initialJmxClient.getPort()) {
                // NOP: I'm already connected on this container, I reuse the connection
                isAlreadyConnected = true;
            } else {
                // It's another container
                isAlreadyConnected = false;
            }

            PetalsAdministration tmpAdmin = admin;
            try {
                if (!isAlreadyConnected) {
                    tmpAdmin = JMXPetalsAdministrationFactory.getInstance()
                            .newPetalsAdministrationAPI();
                    tmpAdmin.connect(container.getHost(), container.getPorts().get(PortType.JMX).intValue(),
                            container.getJmxUsername(), container.getJmxPassword());
                }
                final List<Artifact> artifacts = tmpAdmin.newArtifactAdministration().listArtifacts();
                for (final Artifact artifact : artifacts) {
                    if (artifact instanceof ServiceAssembly) {
                        container.addServiceAssembly((ServiceAssembly) artifact);
                    } else if (artifact instanceof Component) {
                        container.addComponent((Component) artifact);
                    } else if (artifact instanceof SharedLibrary) {
                        container.addSharedLibrary((SharedLibrary) artifact);
                    } else {
                        throw new ArtifactAdministrationException(
                                String.format("Unknown artifact: %s(%s)", artifact.getName(), artifact.getType()));
                    }
                }
            } catch (final ContainerAdministrationException e) {
                // If we are not able to connect to container different from the initial one, we skip to retrieve JBI
                // artifacts.
                // TODO we should return that fact in a way or another… should we mark the container as unreachable?
            } catch (final ArtifactAdministrationException e) {
                throw new ContainerAdministrationException(e);
            } finally {
                if (!isAlreadyConnected && tmpAdmin != null) {
                    try {
                        tmpAdmin.disconnect();
                    } catch (final NoConnectionException e) {
                        // NOP, because no connection previously exists
                    }
                }
            }
        } catch (final NotRemoteJMXClientException | UnknownHostException e) {
            throw new ContainerAdministrationException(e);
        }
    }

    @Override
    public Properties getServerProperties() throws ContainerAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            return client.getPetalsAdminServiceClient().retrieveServerProperties();
        } catch (final ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (final PetalsAdminServiceErrorException pasee) {
            throw new ContainerAdministrationException(pasee);
        } catch (final PetalsAdminDoesNotExistException padnee) {
            throw new ContainerAdministrationException(padnee);
        } catch (final PetalsException e) {
            throw new ContainerAdministrationException(e);
        }
    }

    @Override
    public void changeLoggerLevel(String loggerName, String loggerLevel)
            throws LoggerNotFoundException, InvalidLogLevelException,
            ContainerAdministrationException {
        // TODO: LoggerNotFoundException and InvalidLogLevelException must be
        // thrown here according to the result of setLevelForLogger because the
        // container knows its log level and loggers.
        try {
            JMXClient client = admin.getJMXClient();
            client.getLoggerServiceClient().setLoggerLevel(loggerName, loggerLevel);
        } catch (ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (LoggerServiceErrorException lsee) {
            throw new ContainerAdministrationException(lsee);
        } catch (LoggerServiceDoesNotExistException ldnee) {
            throw new ContainerAdministrationException(ldnee);
        }
    }
    
    @Override
    public List<Logger> getLoggers()
            throws ContainerAdministrationException {
        try {
            JMXClient client = admin.getJMXClient();
            final String[][] loggersArray = client.getLoggerServiceClient().getLoggers();
            final List<Logger> loggers = new LinkedList<>();
            for (int i = 0; i < loggersArray.length; i++) {
                loggers.add(new Logger(loggersArray[i][0], loggersArray[i][1]));
            }
            return loggers;
        } catch (final ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (final LoggerServiceErrorException lsee) {
            throw new ContainerAdministrationException(lsee);
        } catch (final LoggerServiceDoesNotExistException lsdnee) {
            throw new ContainerAdministrationException(lsdnee);
        } catch (final LoggingServiceException e) {
            throw new ContainerAdministrationException(e);
        }
    }

    @Override
    public void attachContainer(final String targetDomainName, final String targetContainerHost,
            final int targetContainerPort, final String targetContainerUser, final String targetContainerPwd,
            final String securityPassPhrase) throws ContainerAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final RemoteContainer targetContainerConfiguration = new RemoteContainer();
            targetContainerConfiguration.setHost(targetContainerHost);
            targetContainerConfiguration.setPort(targetContainerPort);
            targetContainerConfiguration.setUser(targetContainerUser);
            targetContainerConfiguration.setPassword(targetContainerPwd);
            targetContainerConfiguration.setDomainName(targetDomainName);

            client.getTopologyServiceClient().attachContainerTo(targetContainerConfiguration, securityPassPhrase);

        } catch (final ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (final TopologyServiceErrorException lsee) {
            throw new ContainerAdministrationException(lsee);
        } catch (final TopologyServiceDoesNotExistException ldnee) {
            throw new ContainerAdministrationException(ldnee);
        } catch (final TopologyServiceException e) {
            throw new ContainerAdministrationException(e);
        }
    }

    @Override
    public void detachContainer() throws ContainerAdministrationException {
        try {
            admin.getJMXClient().getTopologyServiceClient().detachContainer();

        } catch (final ConnectionErrorException cee) {
            throw new ContainerAdministrationException(cee);
        } catch (final TopologyServiceErrorException lsee) {
            throw new ContainerAdministrationException(lsee);
        } catch (final TopologyServiceDoesNotExistException ldnee) {
            throw new ContainerAdministrationException(ldnee);
        } catch (final TopologyServiceException e) {
            throw new ContainerAdministrationException(e);
        }
    }
}
