/**
 * Copyright (c) 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.jmx;

import org.ow2.petals.admin.api.ArtifactAdministration;
import org.ow2.petals.admin.api.ContainerAdministration;
import org.ow2.petals.admin.api.EndpointDirectoryAdministration;
import org.ow2.petals.admin.api.PetalsAdministration;
import org.ow2.petals.admin.api.artifact.lifecycle.ArtifactLifecycleFactory;
import org.ow2.petals.admin.api.conf.ArtifactUrlRewriter;
import org.ow2.petals.admin.api.exception.ConnectionFailedException;
import org.ow2.petals.admin.api.exception.ContainerAdministrationException;
import org.ow2.petals.admin.api.exception.NoConnectionException;
import org.ow2.petals.admin.jmx.artifact.JMXArtifactLifecycleFactory;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.PetalsJmxApiFactory;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.DuplicatedServiceException;
import org.ow2.petals.jmx.api.api.exception.MissingServiceException;

public class JMXPetalsAdministration implements PetalsAdministration {

    private JMXClient jmxClient;

    /**
     * Create a new JMX client connected to the remote Petals ESB. If a connection already exists, it is closed and a
     * new connection with the specified parameters is created.
     * 
     * @param host
     *            the host of the Petals ESB JMX server
     * @param port
     *            the port of the Petals ESB JMX server
     * @param user
     *            the username of the Petals ESB JMX server
     * @param password
     *            the password of the Petals ESB JMX server
     * 
     * @return the JMX client to communicate with the remote Petals ESB
     * 
     * @throws ContainerAdministrationException
     *             if the client cannot connect to the remote petals ESB
     */
    @Override
    public void connect(String host, int port, String user, String password) throws ContainerAdministrationException {
        try {
            if (jmxClient != null) {
                disconnect();
            }

            jmxClient = PetalsJmxApiFactory.getInstance().createJMXClient(host, port, user, password);
        } catch (final ConnectionErrorException e) {
            throw new ConnectionFailedException(e);
        } catch (final DuplicatedServiceException | MissingServiceException e) {
            throw new ContainerAdministrationException(e);
        }
    }

    /**
     * @return <code>true</code> if a {@link JMXClient} exists (ie. a connection is established to a Petals ESB
     *         container, <code>false</code> otherwise.
     */
    @Override
    public boolean isConnected() throws ContainerAdministrationException {
        return jmxClient != null;
    }

    /**
     * Get the JMX current client connected to the remote petals ESB. This method must be called after the creation of
     * the JMX client.
     * 
     * @return the current JMX client connected to the remote Petals ESB
     * 
     * @throws NoConnectionException
     *             if there is no current client connection to the remote petals ESB
     */
    public final JMXClient getJMXClient() throws NoConnectionException {
        if (jmxClient == null) {
            throw new NoConnectionException();
        }

        return jmxClient;
    }

    /**
     * Disconnect the JMX client.
     * 
     * @throws ConnectionErrorException
     *             An error occurs disconnecting the JMX Client.
     * @throws NoConnectionException
     *             There is no current JMX connection to the remote petals ESB.
     */
    @Override
    public void disconnect() throws NoConnectionException, ContainerAdministrationException {
        try {
            if (jmxClient == null) {
                throw new NoConnectionException();
            }

            jmxClient.disconnect();
            jmxClient = null;
        } catch (ConnectionErrorException cee) {
            // TODO we should maybe only log errors instead of failing?
            throw new ContainerAdministrationException(cee);
        }
    }

    @Override
    public ArtifactAdministration newArtifactAdministration() {
        return new JMXArtifactAdministration(this, ArtifactUrlRewriter.NOOP);
    }

    @Override
    public ArtifactAdministration newArtifactAdministration(final ArtifactUrlRewriter artifactUrlRewriter) {
        return new JMXArtifactAdministration(this, artifactUrlRewriter);
    }

    @Override
    public ContainerAdministration newContainerAdministration() {
        return new JMXContainerAdministration(this);
    }

    @Override
    public EndpointDirectoryAdministration newEndpointDirectoryAdministration() {
        return new JMXEndpointDirectoryAdministration(this);
    }

    @Override
    public ArtifactLifecycleFactory newArtifactLifecycleFactory(final ArtifactUrlRewriter artifactUrlRewriter) {
        return new JMXArtifactLifecycleFactory(this, artifactUrlRewriter);
    }

    @Override
    public ArtifactLifecycleFactory newArtifactLifecycleFactory() {
        return new JMXArtifactLifecycleFactory(this, ArtifactUrlRewriter.NOOP);
    }
}
