/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.jmx.artifact;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.MBeanAttributeInfo;

import org.ow2.petals.admin.api.artifact.ArtifactState.State;
import org.ow2.petals.admin.api.artifact.Component;
import org.ow2.petals.admin.api.artifact.exception.ConfigurationParamComputeException;
import org.ow2.petals.admin.api.artifact.lifecycle.ComponentLifecycle;
import org.ow2.petals.admin.api.conf.ArtifactUrlRewriter;
import org.ow2.petals.admin.api.exception.ArtifactAdministrationException;
import org.ow2.petals.admin.api.exception.ArtifactDeployedException;
import org.ow2.petals.admin.api.exception.ArtifactNotDeployedException;
import org.ow2.petals.admin.api.exception.ArtifactNotFoundException;
import org.ow2.petals.admin.api.exception.ArtifactStartedException;
import org.ow2.petals.admin.api.exception.ArtifactStoppedException;
import org.ow2.petals.admin.api.exception.ArtifactUnsupportedOperationException;
import org.ow2.petals.admin.api.exception.ArtifactUrlRewriterException;
import org.ow2.petals.admin.api.exception.NoConnectionException;
import org.ow2.petals.admin.jmx.JMXPetalsAdministration;
import org.ow2.petals.jmx.api.api.ComponentClient;
import org.ow2.petals.jmx.api.api.InstallationServiceClient;
import org.ow2.petals.jmx.api.api.InstallerComponentClient;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.configuration.component.AbstractConfigurationComponentClient;
import org.ow2.petals.jmx.api.api.configuration.component.InstallationConfigurationComponentClient;
import org.ow2.petals.jmx.api.api.configuration.component.exception.ConfigurationComponentDoesNotExistException;
import org.ow2.petals.jmx.api.api.configuration.component.exception.ConfigurationComponentErrorException;
import org.ow2.petals.jmx.api.api.configuration.component.utils.AttributesHelper;
import org.ow2.petals.jmx.api.api.exception.AttributeErrorException;
import org.ow2.petals.jmx.api.api.exception.ComponentDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.ComponentErrorException;
import org.ow2.petals.jmx.api.api.exception.ComponentUnsupportedOperationException;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.InstallationServiceDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.InstallationServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.InstallerComponentDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.InstallerComponentErrorException;

/**
 * This class represents a remote JBI Component.
 * 
 * @author Nicolas Oddoux - EBM WebSourcing
 */
public class JMXComponentLifecycle extends AbstractLifecycle implements ComponentLifecycle {

    private static final Logger LOG = Logger.getLogger(JMXComponentLifecycle.class.getName());

    private final Component component;

    protected JMXComponentLifecycle(final Component c, final JMXPetalsAdministration admin,
            final ArtifactUrlRewriter artifactUrlRewriter) {
        super(admin, artifactUrlRewriter);
        this.component = c;
    }

    @Override
    public Component getComponent() {
        return this.component;
    }

    @Override
    public void deploy(URL artifactUrl) throws ArtifactDeployedException, ArtifactAdministrationException {
        this.deploy(artifactUrl, false);
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for component deployment:
     * <ul>
     * <li>if the component installer is already loaded, it will be unloaded
     * before to reload a new one.</li>
     * <li>if the component is already installed and in state 'SHUTDOWN', it
     * will be uninstalled before to unload the component installer and reload a
     * new one,</li>
     * <li>otherwise an error is thrown.</li>
     * </ul>
     * </p>
     */
    @Override
    public void deploy(final URL artifactUrl, final boolean skipInstall) throws ArtifactDeployedException,
            ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String componentName = this.getComponent().getName();
            final InstallationServiceClient installSvcClt = client.getInstallationServiceClient();

            InstallerComponentClient installer;
            try {
                installer = installSvcClt.loadInstaller(componentName);
                // A previous installer already exist:
                // - if the component is already installed and in state
                // 'SHUTDOWN', we uninstall it before to unload the component
                // installer and reload a new one,
                // - if the component is not already installed, we unload the
                // component installer and reload a new one,
                // - otherwise an error is thrown
                try {
                    final ComponentClient componentCltSvc = client
                            .getComponentClient(componentName);
                    final ComponentClient.State state = componentCltSvc.getCurrentState();
                    if (state == ComponentClient.State.SHUTDOWN || state == ComponentClient.State.UNKNOWN) {
                        // The component is installed and in state 'SHUTDOWN'
                        LOG.warning("The component is in state '"
                                + state
                                + "', it has been automatically uninstalled and unloaded.");
                        installer.uninstall();
                    } else {
                        throw new ArtifactDeployedException(componentName, state.toString());
                    }
                } catch (final ComponentDoesNotExistException e) {
                    // The component is not already installed
                    LOG.warning("The component installer is already loaded and only loaded, it has been automatically unloaded.");
                }

                if (installSvcClt.unloadInstaller(componentName)) {
                    installer = installSvcClt.loadNewInstaller(artifactUrlRewriter.rewrite(artifactUrl));
                } else {
                    throw new ArtifactAdministrationException(
                            "Impossible to unload the previously existing component installer");
                }
            } catch (final InstallerComponentDoesNotExistException e) {
                installer = installSvcClt.loadNewInstaller(artifactUrlRewriter.rewrite(artifactUrl));
            }

            if (!skipInstall) {
                InstallationConfigurationComponentClient icc = null;
                try {
                    icc = installer.getInstallationConfigurationClient();
                } catch (ConfigurationComponentDoesNotExistException e) {
                    // handled in the method below
                }

                setParameters(icc);

                installer.install();
            }
        } catch (final ConnectionErrorException | InstallerComponentErrorException | ArtifactUrlRewriterException
                | ComponentErrorException | InstallationServiceErrorException | ConfigurationComponentErrorException
                | InstallationServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    @Override
    public void install() throws ArtifactDeployedException, ArtifactNotFoundException, ArtifactAdministrationException {
        final String componentName = this.getComponent().getName();

        try {
            final JMXClient client = admin.getJMXClient();
            final InstallationServiceClient installSvcClt = client.getInstallationServiceClient();
            final InstallerComponentClient installer = installSvcClt.loadInstaller(componentName);

            // let's check that it is not already installed!
            try {
                final ComponentClient componentCltSvc = client.getComponentClient(componentName);
                // it's installed, let's stop there
                throw new ArtifactDeployedException(componentName, componentCltSvc.getCurrentState().toString());
            } catch (ComponentDoesNotExistException e) {
                // it's not installed, good!
            }

            installer.install();

        } catch (final ConnectionErrorException | InstallationServiceDoesNotExistException
                | InstallationServiceErrorException | ComponentErrorException | InstallerComponentErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (InstallerComponentDoesNotExistException e) {
            throw new ArtifactNotFoundException(this.getComponent().getType(), componentName,
                    this.getComponent().getVersion());
        }
    }

    @Override
    public void uninstall()
            throws ArtifactNotDeployedException, ArtifactStartedException, ArtifactNotFoundException,
            ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();

            uninstall(client);

        } catch (final NoConnectionException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    private void uninstall(final JMXClient client) throws ArtifactNotFoundException, ArtifactNotDeployedException,
            ArtifactStartedException, ArtifactAdministrationException {
        String componentName = this.getComponent().getName();
        try {
            final InstallationServiceClient installSvcClt = client.getInstallationServiceClient();
            final InstallerComponentClient installer = installSvcClt.loadInstaller(componentName);

            // let's check that it is installed
            final ComponentClient componentCltSvc = client.getComponentClient(componentName);

            // let's check that it is not started
            final ComponentClient.State state = componentCltSvc.getCurrentState();
            if (state == ComponentClient.State.STARTED) {
                throw new ArtifactStartedException(componentName, this.getComponent().getType());
            } else if (state == ComponentClient.State.STOPPED) {
                componentCltSvc.shutdown();
            }

            installer.uninstall();
        } catch (final ConnectionErrorException | InstallationServiceDoesNotExistException
                | InstallationServiceErrorException | ComponentErrorException | InstallerComponentErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (InstallerComponentDoesNotExistException e) {
            throw new ArtifactNotFoundException(this.getComponent().getType(), componentName,
                    this.getComponent().getVersion());
        } catch (ComponentDoesNotExistException e) {
            throw new ArtifactNotDeployedException(this.getComponent().getType(), componentName);
        }
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for component start:
     * <ul>
     * <li>if the component is in state 'STARTED', the error
     * {@link ArtifactStartedException} is raised,</li>
     * <li>if the component is in state 'STOPPED', it will be started,</li>
     * <li>if the component is in state 'SHUTDOWN', it will be started,</li>
     * <li>if the component is not installed but component installer is loaded,
     * the error {@link ArtifactNotDeployedException} is raised,</li>
     * <li>if the installer is not loaded, the error
     * {@link ArtifactNotFoundException} is raised.</li>
     * </ul>
     * </p>
     */
    @Override
    public void start() throws ArtifactStartedException, ArtifactNotDeployedException,
            ArtifactNotFoundException, ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String componentName = this.getComponent().getName();

            try {
                final ComponentClient componentCltSvc = client.getComponentClient(componentName);
                final ComponentClient.State state = componentCltSvc.getCurrentState();
                if (state == ComponentClient.State.STARTED) {
                    // The component is installed and its state is 'STARTED'
                    throw new ArtifactStartedException(componentName, this.getComponent().getType());
                }

                componentCltSvc.start();

            } catch (final ComponentDoesNotExistException e) {
                try {
                    // The component is not already installed, we check if the
                    // component installer is loaded
                    client.getInstallationServiceClient().loadInstaller(componentName);
                    // Here, we have checked that the component installer was
                    // loaded
                    throw new ArtifactNotDeployedException(this.getComponent().getType(),
                            componentName);
                } catch (final InstallerComponentDoesNotExistException e1) {
                    // The component installer is not loaded
                    throw new ArtifactNotFoundException(this.getComponent().getType(), componentName, this
                            .getComponent().getVersion());
                }
            }
        } catch (ConnectionErrorException | InstallationServiceErrorException | InstallationServiceDoesNotExistException
                | ComponentErrorException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for component stop:
     * <ul>
     * <li>if the component is in state 'STARTED', it will be stopped,</li>
     * <li>if the component is in state 'STOPPED', the error
     * {@link ArtifactStoppedException} is raised,</li>
     * <li>if the component is in state 'SHUTDOWN', the error
     * {@link ArtifactStoppedException} is raised,</li>
     * <li>if the component is not installed but component installer is loaded,
     * the error {@link ArtifactNotDeployedException} is raised,</li>
     * <li>if the installer is not loaded, the error
     * {@link ArtifactNotFoundException} is raised.</li>
     * </ul>
     * </p>
     */
    @Override
    public void stop() throws ArtifactStoppedException, ArtifactNotDeployedException,
            ArtifactNotFoundException, ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String componentName = this.getComponent().getName();
            try {
                final ComponentClient componentCltSvc = client
                        .getComponentClient(componentName);
                final ComponentClient.State state = componentCltSvc.getCurrentState();
                if (state == ComponentClient.State.STOPPED || state == ComponentClient.State.SHUTDOWN ) {
                    // The component is installed and its state is 'STOPPED' or 'SHUTDOWN'
                    throw new ArtifactStoppedException(componentName, this.getComponent()
                            .getType());
                }
                
                componentCltSvc.stop();
                
            } catch (final ComponentDoesNotExistException e) {
                try {
                    // The component is not already installed, we check if the component installer is loaded
                    client.getInstallationServiceClient().loadInstaller(componentName);
                    // Here, we have checked that the component installer was loaded
                    throw new ArtifactNotDeployedException(this.getComponent().getType(),
                            componentName);
                } catch (final InstallerComponentDoesNotExistException e1) {
                    // The component installer is not loaded
                    throw new ArtifactNotFoundException(this.getComponent().getType(), componentName, this
                            .getComponent().getVersion());
                }
            }
            
        } catch (final ConnectionErrorException | InstallationServiceErrorException
                | InstallationServiceDoesNotExistException | ComponentErrorException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for component undeployment:
     * <ul>
     * <li>if the component is in state 'STARTED', the error
     * {@link ArtifactStartedException} is raised,</li>
     * <li>if the component is in state 'STOPPED', the component will be
     * shutdown, uninstalled and the component installer will be unloaded,</li>
     * <li>if the component is in state 'SHUTDOWN', the component will be
     * uninstalled and the component installer will be unloaded,</li>
     * <li>if the component is not installed but component installer is loaded,
     * the component installer will be unloaded,</li>
     * <li>if the installer is not loaded, the error
     * {@link ArtifactNotFoundException} is raised.</li>
     * </ul>
     * </p>
     */
    @Override
    public void undeploy() throws ArtifactStartedException, ArtifactNotFoundException,
            ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String componentName = this.getComponent().getName();
            final InstallationServiceClient installSvcClt = client.getInstallationServiceClient();

            try {
                uninstall(client);
            } catch (final ArtifactNotDeployedException e) {
                // it's loaded because if not we would have directly got an ArtifactNotFoundException
                LOG.warning("The component is not installed but loaded, it will be automatically unloaded.");
            }

            // Here, the component installer should be unload
            final boolean isSuccessful = installSvcClt.unloadInstaller(componentName);
            if (!isSuccessful) {
                throw new ArtifactAdministrationException(
                        String.format("Failed to unload the component '%s'", componentName));
            }

        } catch (final ConnectionErrorException | InstallationServiceErrorException
                | InstallationServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    @Override
    public void updateStateAndParameters() throws ArtifactNotFoundException, ArtifactAdministrationException {
        update(true);
    }

    @Override
    public void updateState() throws ArtifactNotFoundException, ArtifactAdministrationException {
        update(false);
    }

    private void update(boolean withParameters) throws ArtifactNotFoundException, ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String name = this.getComponent().getName();

            State state;
            AbstractConfigurationComponentClient ccc = null;
            try {
                ComponentClient cc = client.getComponentClient(name);
                org.ow2.petals.jmx.api.api.ComponentClient.State currentState = cc.getCurrentState();
                state = currentState == null ? State.UNKNOWN : State.fromStringValue(currentState.toString());
                if (withParameters) {
                    try {
                        ccc = cc.getRuntimeConfigurationClient();
                    } catch (final ConfigurationComponentDoesNotExistException e) {
                        // it's ok, there is no parameters thus
                    }
                }
            } catch (final ComponentDoesNotExistException e) {
                // could be only installed
                try {
                    InstallerComponentClient icc = client.getInstallationServiceClient().loadInstaller(name);
                    state = State.LOADED;
                    if (withParameters) {
                        try {
                            ccc = icc.getInstallationConfigurationClient();
                        } catch (ConfigurationComponentDoesNotExistException e2) {
                            // it's ok, there is no parameters thus
                        }
                    }
                } catch (InstallerComponentDoesNotExistException e2) {
                    throw new ArtifactNotFoundException(this.getComponent().getType(), name, null);
                }
            }

            this.getComponent().setState(state);

            if (withParameters) {
                this.getComponent().getParameters().clear();
                if (ccc != null) {
                    for (Entry<MBeanAttributeInfo, Object> entry : ccc.getConfigurationMBeanAttributes().entrySet()) {
                        this.getComponent().getParameters().setProperty(entry.getKey().getName(),
                                Objects.toString(entry.getValue()));
                    }
                }
            }
        } catch (final ConnectionErrorException | ComponentErrorException | InstallerComponentErrorException
                | ConfigurationComponentErrorException | InstallationServiceErrorException
                | InstallationServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    private void setParameters(AbstractConfigurationComponentClient ccc) throws ArtifactAdministrationException {
        try {
            if (ccc != null) {
                final Map<MBeanAttributeInfo, Object> mbeanAttributes = ccc.getConfigurationMBeanAttributes();
                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine("Available configuration parameters at component level: " + mbeanAttributes.toString());
                }
                final Map<String, String> existingConfParams = new HashMap<>();
                for (final Entry<MBeanAttributeInfo, Object> mbeanAttribute : mbeanAttributes.entrySet()) {
                    final String paramName = mbeanAttribute.getKey().getName();
                    final Object paramValueObj = mbeanAttribute.getValue();
                    if (LOG.isLoggable(Level.FINE)) {
                        LOG.fine(String.format("Setting '%s' to '%s'", paramName, paramValueObj));
                    }
                    existingConfParams.put(paramName, paramValueObj == null ? null : paramValueObj.toString());
                }

                final Properties parameters = this.getComponent().configure(existingConfParams);

                if (parameters != null) {
                    for (String parameterName : parameters.stringPropertyNames()) {
                        Object parameterValue = AttributesHelper.convertParameterValueToObject(parameterName,
                                parameters.getProperty(parameterName), mbeanAttributes);
                        ccc.setAttribute(parameterName, parameterValue);
                    }
                }
            } else {
                // If no configuration MBean is available, and configuration parameters were given to be
                // set, an error must occur
                if (!this.getComponent().getParameters().isEmpty()) {
                    throw new ArtifactAdministrationException(
                            "Impossible to set configuration parameters for this component (no configuration MBean)");
                }
            }
        } catch (ConfigurationComponentErrorException | AttributeErrorException
                | ConfigurationParamComputeException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    @Override
    public void setParameters() throws ArtifactNotFoundException, ArtifactAdministrationException {
        final JMXClient client = admin.getJMXClient();
        final String name = this.getComponent().getName();
        try {
            InstallerComponentClient installer = client.getInstallationServiceClient().loadInstaller(name);

            AbstractConfigurationComponentClient ccc = null;
            try {
                if (installer.isInstalled()) {
                    ccc = client.getComponentClient(name).getRuntimeConfigurationClient();
                } else {
                    ccc = installer.getInstallationConfigurationClient();
                }
            } catch (ConfigurationComponentDoesNotExistException e) {
                // handled in method
            }

            setParameters(ccc);
        } catch (InstallationServiceErrorException | InstallerComponentErrorException
                | ConfigurationComponentErrorException | InstallationServiceDoesNotExistException
                | ConnectionErrorException | ComponentErrorException e) {
            throw new ArtifactAdministrationException(e);
        } catch (ComponentDoesNotExistException | InstallerComponentDoesNotExistException e) {
            throw new ArtifactNotFoundException(this.getComponent().getType(), name, null);
        }
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for component reload configuration:
     * <ul>
     * <li>if the component is not installed but component installer is loaded, the error
     * {@link ArtifactNotFoundException} is raised,</li>
     * <li>if the installer is not loaded, the error {@link ArtifactNotFoundException} is raised.</li>
     * </ul>
     * </p>
     */
    @Override
    public void reloadConfiguration() throws ArtifactNotFoundException, ArtifactUnsupportedOperationException,
            ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String componentName = this.getComponent().getName();
            try {
                final ComponentClient componentCltSvc = client.getComponentClient(componentName);

                componentCltSvc.reloadConfiguration();

            } catch (final ComponentDoesNotExistException e) {
                throw new ArtifactNotFoundException(this.getComponent().getType(), componentName, this.getComponent()
                        .getVersion());
            } catch (final ComponentUnsupportedOperationException e) {
                throw new ArtifactUnsupportedOperationException();
            }

        } catch (final ConnectionErrorException | ComponentErrorException e) {
            throw new ArtifactAdministrationException(e);
        }
    }
}
