/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.jmx.artifact;

import java.net.URL;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.ow2.petals.admin.api.artifact.ArtifactState.State;
import org.ow2.petals.admin.api.artifact.ServiceAssembly;
import org.ow2.petals.admin.api.artifact.lifecycle.ServiceAssemblyLifecycle;
import org.ow2.petals.admin.api.conf.ArtifactUrlRewriter;
import org.ow2.petals.admin.api.exception.ArtifactAdministrationException;
import org.ow2.petals.admin.api.exception.ArtifactDeployedException;
import org.ow2.petals.admin.api.exception.ArtifactNotDeployedException;
import org.ow2.petals.admin.api.exception.ArtifactNotFoundException;
import org.ow2.petals.admin.api.exception.ArtifactStartedException;
import org.ow2.petals.admin.api.exception.ArtifactStoppedException;
import org.ow2.petals.admin.api.exception.ArtifactUrlRewriterException;
import org.ow2.petals.admin.jmx.JMXPetalsAdministration;
import org.ow2.petals.jmx.api.api.DeploymentServiceClient;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.DeploymentServiceDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.DeploymentServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.ServiceAssemblyDoesNotExistException;

/**
 * This class represents a remote JBI service assembly.
 * 
 * @author Nicolas Oddoux - EBM WebSourcing
 */
public class JMXServiceAssemblyLifecycle extends AbstractLifecycle implements ServiceAssemblyLifecycle {

    private static final Logger LOG = Logger.getLogger(JMXServiceAssemblyLifecycle.class.getName());

    private static final String SA_SUCCESS_RESULT = "SUCCESS";

    private static final Pattern SA_RESULT_PATTERN = Pattern
            .compile(
                    ".+?<frmwk-task-result>.+?<task-result>(\\w+?)</task-result>.+?<loc-param>(.+?)</loc-param>",
                    Pattern.DOTALL);

    private static final Pattern SU_RESULT_PATTERN = Pattern
            .compile(
                    ".+?<component-task-result>.+?<task-result>FAILED</task-result>.+?<loc-param>(.+?)</loc-param>.+?<stack-trace>(.+?)</stack-trace>",
                    Pattern.DOTALL);

    private final ServiceAssembly serviceAssembly;

    protected JMXServiceAssemblyLifecycle(final ServiceAssembly sa, final JMXPetalsAdministration admin,
            final ArtifactUrlRewriter artifactUrlRewriter) {
        super(admin, artifactUrlRewriter);
        this.serviceAssembly = sa;
    }

    @Override
    public ServiceAssembly getServiceAssembly() {
        return serviceAssembly;
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for service assembly deployment:
     * <ul>
     * <li>if the service assembly is in state 'SHUTDOWN', it will be undeployed
     * before to be be redeplyed,</li>
     * <li>otherwise an error is thrown.</li>
     * </ul>
     * </p>
     */
    @Override
    public void deploy(URL artifactUrl) throws ArtifactDeployedException,
            ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String saName = this.getServiceAssembly().getName();
            final DeploymentServiceClient dplSvcClt = client.getDeploymentServiceClient();
            try {
                final DeploymentServiceClient.State state = dplSvcClt.getState(saName);
                if (state != DeploymentServiceClient.State.SHUTDOWN && state != DeploymentServiceClient.State.UNKNOWN) {
                    throw new ArtifactDeployedException(saName, state.toString());
                } else {
                    LOG.warning("The service assembly is in state '"
                            + state
                            + "', it has been automatically undeployed before to be re-deployed.");
                    checkResult(client.getDeploymentServiceClient().undeploy(saName), saName,
                            "Failed to undeploy SA '%s'");
                }

            } catch (final ServiceAssemblyDoesNotExistException e) {
                // The service assembly is not already deployed
            }

            checkResult(
                    client.getDeploymentServiceClient().deploy(artifactUrlRewriter.rewrite(artifactUrl)),
                    "Failed to deploy SU '%s':\n%s",
                    "SA '%s' deployed with some SU deployment in failure",
                    "Failed to deploy SA '%s'");

        } catch (final ConnectionErrorException | DeploymentServiceErrorException
                | DeploymentServiceDoesNotExistException | ArtifactUrlRewriterException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for service assembly start:
     * <ul>
     * <li>if the SA is in state 'STARTED', the error
     * {@link ArtifactStartedException} is raised,</li>
     * <li>if the SA is in state 'STOPPED', it will be started,</li>
     * <li>if the SA is in state 'SHUTDOWN', it will be started,</li>
     * <li>if the SA is not deployed, the error
     * {@link ArtifactNotFoundException} is raised.</li>
     * </ul>
     * </p>
     */
    @Override
    public void start() throws ArtifactStartedException, ArtifactNotDeployedException,
            ArtifactNotFoundException, ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String saName = this.getServiceAssembly().getName();
            final DeploymentServiceClient dplSvcClt = client.getDeploymentServiceClient();
            try {
                final DeploymentServiceClient.State state = dplSvcClt.getState(saName);
                if (state == DeploymentServiceClient.State.STARTED) {
                    throw new ArtifactStartedException(this.getServiceAssembly().getType(), saName);
                } else {
                    checkResult(client.getDeploymentServiceClient().start(saName), saName,
                            "Failed to start SA '%s'");
                }
            } catch (final ServiceAssemblyDoesNotExistException e) {
                throw new ArtifactNotFoundException(this.getServiceAssembly().getType(), saName, this
                        .getServiceAssembly().getVersion());
            }
        } catch (ConnectionErrorException | DeploymentServiceErrorException
                | DeploymentServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for service assembly stop:
     * <ul>
     * <li>if the SA is in state 'STARTED', it will be stop,</li>
     * <li>if the SA is in state 'STOPPED', the error
     * {@link ArtifactStoppedException} is raised,</li>
     * <li>if the SA is in state 'SHUTDOWN', the error
     * {@link ArtifactStoppedException} is raised,</li>
     * <li>if the SA is not deployed, the error
     * {@link ArtifactNotFoundException} is raised.</li>
     * </ul>
     * </p>
     */
    @Override
    public void stop() throws ArtifactStoppedException, ArtifactNotFoundException,
            ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String saName = this.getServiceAssembly().getName();
            final DeploymentServiceClient dplSvcClt = client.getDeploymentServiceClient();
            try {
                final DeploymentServiceClient.State state = dplSvcClt.getState(saName);
                if (state == DeploymentServiceClient.State.STOPPED
                        || state == DeploymentServiceClient.State.SHUTDOWN) {
                    throw new ArtifactStoppedException(this.getServiceAssembly().getType(), saName);
                } else {
                    checkResult(client.getDeploymentServiceClient().stop(saName), saName,
                            "Failed to stop SA '%s'");
                }
            } catch (final ServiceAssemblyDoesNotExistException e) {
                throw new ArtifactNotFoundException(this.getServiceAssembly().getType(), saName, this
                        .getServiceAssembly().getVersion());
            }
        } catch (final ConnectionErrorException | DeploymentServiceErrorException
                | DeploymentServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    /**
     * {@inheritDoc}
     * <p>
     * Special note for service assembly undeployment:
     * <ul>
     * <li>if the SA is in state 'STARTED', the error
     * {@link ArtifactStartedException} is raised,</li>
     * <li>if the SA is in state 'STOPPED', the SA will be shutdown and
     * undeployed,</li>
     * <li>if the SA is in state 'SHUTDOWN', the SA will be undeployed.</li>
     * <li>if the SA is not deployed, the error
     * {@link ArtifactNotFoundException} is raised.</li>
     * </ul>
     * </p>
     */
    @Override
    public void undeploy() throws ArtifactStartedException, ArtifactNotFoundException,
            ArtifactAdministrationException {
        try {
            final JMXClient client = admin.getJMXClient();
            final String saName = this.getServiceAssembly().getName();
            final DeploymentServiceClient dplSvcClt = client.getDeploymentServiceClient();
            try {
                final DeploymentServiceClient.State state = dplSvcClt.getState(saName);

                if (state == DeploymentServiceClient.State.STARTED) {
                    throw new ArtifactStartedException(this.getServiceAssembly().getType(), saName);
                }

                if (state == DeploymentServiceClient.State.STOPPED) {
                    checkResult(dplSvcClt.shutdown(saName), saName, "Failed to shutdown SA '%s'");
                }

                checkResult(dplSvcClt.undeploy(saName), "Failed to undeploy SU '%s':\n%s",
                        "SA '%s' undeployed with some SU undeployment in failure", "Failed to undeploy SA '%s'");
            } catch (final ServiceAssemblyDoesNotExistException e) {
                throw new ArtifactNotFoundException(this.getServiceAssembly().getType(), saName, this
                        .getServiceAssembly().getVersion());
            }
        } catch (ConnectionErrorException | DeploymentServiceErrorException
                | DeploymentServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        }
    }

    private static final void checkResult(String serviceAssemblyReport,
            String suErrorMessageFormat, String saPartialErrorMessageFormat,
            String saErrorMessageFormat) throws ArtifactAdministrationException {
        boolean isFailed = false;
        StringBuilder errorMessage = new StringBuilder();
        Matcher matcher = SU_RESULT_PATTERN.matcher(serviceAssemblyReport);
        if (matcher.find()) {
            isFailed = true;
            errorMessage.append(String.format(suErrorMessageFormat, matcher.group(1),
                    matcher.group(2)));
        }
        matcher = SA_RESULT_PATTERN.matcher(serviceAssemblyReport);
        if (matcher.find()) {
            if (SA_SUCCESS_RESULT.equals(matcher.group(1).toUpperCase())) {
                if (isFailed) {
                    errorMessage
                            .append(String.format(saPartialErrorMessageFormat, matcher.group(2)));
                }
            } else {
                isFailed = true;
                errorMessage.append(String.format(saErrorMessageFormat, matcher.group(2)));
            }
        }

        if (isFailed) {
            throw new ArtifactAdministrationException(errorMessage.toString());
        }
    }

    private static final void checkResult(String serviceAssemblyReport, String errorMessageFormat,
            String saName) throws ArtifactAdministrationException {
        final Matcher matcher = SA_RESULT_PATTERN.matcher(serviceAssemblyReport);
        if (matcher.find()) {
            if (!SA_SUCCESS_RESULT.equals(matcher.group(1).toUpperCase())) {
                String errorMessage = String.format(errorMessageFormat, saName);
                throw new ArtifactAdministrationException(errorMessage);
            }
        }
    }

    @Override
    public void updateState() throws ArtifactNotFoundException, ArtifactAdministrationException {
        final String serviceAssemblyName = this.getServiceAssembly().getName();
        try {
            JMXClient client = admin.getJMXClient();
            final DeploymentServiceClient.State state = client.getDeploymentServiceClient()
                    .getState(serviceAssemblyName);
            this.getServiceAssembly().setState(State.fromStringValue(state.toString()));
        } catch (final ConnectionErrorException | DeploymentServiceErrorException
                | DeploymentServiceDoesNotExistException e) {
            throw new ArtifactAdministrationException(e);
        } catch (final ServiceAssemblyDoesNotExistException e) {
            throw new ArtifactNotFoundException(this.getServiceAssembly().getType(), serviceAssemblyName, null);
        }
    }
}
