/**
 * Copyright (c) 2015-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.jmx;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.admin.api.ArtifactAdministration;
import org.ow2.petals.admin.api.PetalsAdministration;
import org.ow2.petals.admin.api.PetalsAdministrationFactory;
import org.ow2.petals.admin.api.artifact.Artifact;
import org.ow2.petals.admin.api.exception.ArtifactAdministrationException;
import org.ow2.petals.admin.api.exception.ArtifactNotFoundException;
import org.ow2.petals.admin.api.exception.ArtifactTypeIsNeededException;
import org.ow2.petals.admin.api.exception.ArtifactVersionIsNeededException;
import org.ow2.petals.admin.api.exception.ContainerAdministrationException;
import org.ow2.petals.admin.api.exception.DuplicatedServiceException;
import org.ow2.petals.admin.api.exception.MissingServiceException;
import org.ow2.petals.admin.api.exception.UnsupportedArtifactTypeException;
import org.ow2.petals.jmx.api.api.exception.InstallationServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.NotRemoteJMXClientException;
import org.ow2.petals.jmx.api.mock.junit.PetalsJmxApiJunitRule;
import org.ow2.petals.jmx.api.mock.junit.SharedLibrary;

/**
 * Unit tests of {@link JMXContainerAdministration}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class JMXArtifactAdministrationTest {

    private static final String SL1_NAME = "sl1-name";

    private static final String SL1_VERSION = "sl1-version";

    private static final String SL2_NAME = "sl2-name";

    private static final String SL2_VERSION = "sl2-version";

    @Rule
    public final PetalsJmxApiJunitRule jmxApi = new PetalsJmxApiJunitRule();

    /**
     * <p>
     * Try to retrieve a SL into the following environment:
     * <ul>
     * <li>a shared library is installed with only one version.</li>
     * </ul>
     * using:
     * <ul>
     * <li>its name,</li>
     * <li>the version is not set,</li>
     * <li>the type is not set.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: No error occurs, the SL is retrieved.
     * </p>
     */
    @Test
    public void getArtifactSLFromNameIntoEnvWithOneVersions() throws InstallationServiceErrorException,
            NotRemoteJMXClientException, ContainerAdministrationException, DuplicatedServiceException,
            MissingServiceException, UnsupportedArtifactTypeException, ArtifactNotFoundException,
            ArtifactTypeIsNeededException, ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");
        final SharedLibrary sl1 = new SharedLibrary(SL1_NAME, SL1_VERSION);
        this.jmxApi.addSharedLibrary(sl1);

        final ArtifactAdministration artifactAdm = containerAdm.newArtifactAdministration();
        assertTrue(artifactAdm instanceof JMXArtifactAdministration);

        final Artifact expectedArtifact = artifactAdm.getArtifact(null, SL1_NAME, null);

        assertNotNull(expectedArtifact);
        assertTrue(expectedArtifact instanceof org.ow2.petals.admin.api.artifact.SharedLibrary);
        assertEquals(SL1_NAME, expectedArtifact.getName());
        assertEquals(SL1_VERSION, expectedArtifact.getVersion());
        assertEquals(org.ow2.petals.admin.api.artifact.SharedLibrary.TYPE, expectedArtifact.getType());
    }

    /**
     * <p>
     * Try to retrieve a SL into the following environment:
     * <ul>
     * <li>a shared library is installed with only one version.</li>
     * </ul>
     * using:
     * <ul>
     * <li>its name,</li>
     * <li>its version,</li>
     * <li>the type is not set.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: No error occurs, the SL is retrieved.
     * </p>
     */
    @Test
    public void getArtifactSLFromNameAndVersionIntoEnvWithOneVersions() throws InstallationServiceErrorException,
            NotRemoteJMXClientException, ContainerAdministrationException, DuplicatedServiceException,
            MissingServiceException, UnsupportedArtifactTypeException, ArtifactNotFoundException,
            ArtifactTypeIsNeededException, ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");
        final SharedLibrary sl1 = new SharedLibrary(SL1_NAME, SL1_VERSION);
        this.jmxApi.addSharedLibrary(sl1);

        final ArtifactAdministration artifactAdm = containerAdm.newArtifactAdministration();
        assertTrue(artifactAdm instanceof JMXArtifactAdministration);

        final Artifact expectedArtifact = artifactAdm.getArtifact(null, SL1_NAME, SL1_VERSION);

        assertNotNull(expectedArtifact);
        assertTrue(expectedArtifact instanceof org.ow2.petals.admin.api.artifact.SharedLibrary);
        assertEquals(SL1_NAME, expectedArtifact.getName());
        assertEquals(SL1_VERSION, expectedArtifact.getVersion());
        assertEquals(org.ow2.petals.admin.api.artifact.SharedLibrary.TYPE, expectedArtifact.getType());
    }

    /**
     * <p>
     * Try to retrieve a SL into the following environment:
     * <ul>
     * <li>a shared library is installed with only one version.</li>
     * </ul>
     * using:
     * <ul>
     * <li>the name is not set,</li>
     * <li>its version,</li>
     * <li>the type is not set.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the missing artifact name
     * </p>
     */
    @Test(expected = ArtifactNotFoundException.class)
    public void getArtifactSLWithoutName() throws InstallationServiceErrorException, NotRemoteJMXClientException,
            ContainerAdministrationException, DuplicatedServiceException, MissingServiceException,
            UnsupportedArtifactTypeException, ArtifactNotFoundException, ArtifactTypeIsNeededException,
            ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");
        final SharedLibrary sl1 = new SharedLibrary(SL1_NAME, SL1_VERSION);
        this.jmxApi.addSharedLibrary(sl1);

        final ArtifactAdministration artifactAdm = containerAdm.newArtifactAdministration();
        assertTrue(artifactAdm instanceof JMXArtifactAdministration);
        artifactAdm.getArtifact(null, null, SL1_VERSION);
    }

    /**
     * <p>
     * Try to retrieve a SL into the following environment:
     * <ul>
     * <li>a shared library is installed with two versions.</li>
     * </ul>
     * using:
     * <ul>
     * <li>its name,</li>
     * <li>its version,</li>
     * <li>the type is not set.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: No error occurs, the SL is retrieved.
     * </p>
     */
    @Test
    public void getArtifactSLFromNameAndVersionIntoEnvWithTwoVersions() throws InstallationServiceErrorException,
            NotRemoteJMXClientException, ContainerAdministrationException, DuplicatedServiceException,
            MissingServiceException, UnsupportedArtifactTypeException, ArtifactNotFoundException,
            ArtifactTypeIsNeededException, ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");
        final SharedLibrary sl1 = new SharedLibrary(SL1_NAME, SL1_VERSION);
        this.jmxApi.addSharedLibrary(sl1);
        final SharedLibrary sl2 = new SharedLibrary(SL1_NAME, SL2_VERSION);
        this.jmxApi.addSharedLibrary(sl2);

        final ArtifactAdministration artifactAdm = containerAdm.newArtifactAdministration();
        assertTrue(artifactAdm instanceof JMXArtifactAdministration);

        final Artifact expectedArtifact = artifactAdm.getArtifact(null, SL1_NAME, SL1_VERSION);

        assertNotNull(expectedArtifact);
        assertTrue(expectedArtifact instanceof org.ow2.petals.admin.api.artifact.SharedLibrary);
        assertEquals(SL1_NAME, expectedArtifact.getName());
        assertEquals(SL1_VERSION, expectedArtifact.getVersion());
        assertEquals(org.ow2.petals.admin.api.artifact.SharedLibrary.TYPE, expectedArtifact.getType());
    }

    /**
     * <p>
     * Try to retrieve a SL into the following environment:
     * <ul>
     * <li>a shared library is installed with two versions.</li>
     * </ul>
     * using:
     * <ul>
     * <li>its name,</li>
     * <li>the version is not set,</li>
     * <li>the type is not set.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: An error occurs about the missing version.
     * </p>
     */
    @Test(expected = ArtifactVersionIsNeededException.class)
    public void getArtifactSLFromNameIntoEnvWithTwoVersions() throws InstallationServiceErrorException,
            NotRemoteJMXClientException, ContainerAdministrationException, DuplicatedServiceException,
            MissingServiceException, UnsupportedArtifactTypeException, ArtifactNotFoundException,
            ArtifactTypeIsNeededException, ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");
        final SharedLibrary sl1 = new SharedLibrary(SL1_NAME, SL1_VERSION);
        this.jmxApi.addSharedLibrary(sl1);
        final SharedLibrary sl2 = new SharedLibrary(SL1_NAME, SL2_VERSION);
        this.jmxApi.addSharedLibrary(sl2);

        final ArtifactAdministration artifactAdm = containerAdm.newArtifactAdministration();
        assertTrue(artifactAdm instanceof JMXArtifactAdministration);

        artifactAdm.getArtifact(null, SL1_NAME, null);
    }

    /**
     * <p>
     * Try to retrieve a SL into the following environment:
     * <ul>
     * <li>two shared libraries are installed, one version per SL.</li>
     * </ul>
     * using:
     * <ul>
     * <li>its name,</li>
     * <li>the version is not set,</li>
     * <li>the type is not set.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: No error occurs, the SL is retrieved.
     * </p>
     */
    @Test
    public void getArtifactSLFromNameIntoEnvWithTwoDifferentSls() throws InstallationServiceErrorException,
            NotRemoteJMXClientException, ContainerAdministrationException, DuplicatedServiceException,
            MissingServiceException, UnsupportedArtifactTypeException, ArtifactNotFoundException,
            ArtifactTypeIsNeededException, ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");
        final SharedLibrary sl1 = new SharedLibrary(SL1_NAME, SL1_VERSION);
        this.jmxApi.addSharedLibrary(sl1);
        final SharedLibrary sl2 = new SharedLibrary(SL2_NAME, SL2_VERSION);
        this.jmxApi.addSharedLibrary(sl2);

        final ArtifactAdministration artifactAdm = containerAdm.newArtifactAdministration();
        assertTrue(artifactAdm instanceof JMXArtifactAdministration);

        final Artifact expectedArtifact = artifactAdm.getArtifact(null, SL1_NAME, null);

        assertNotNull(expectedArtifact);
        assertTrue(expectedArtifact instanceof org.ow2.petals.admin.api.artifact.SharedLibrary);
        assertEquals(SL1_NAME, expectedArtifact.getName());
        assertEquals(SL1_VERSION, expectedArtifact.getVersion());
        assertEquals(org.ow2.petals.admin.api.artifact.SharedLibrary.TYPE, expectedArtifact.getType());
    }

    /**
     * <p>
     * Try to retrieve a SL into the following environment:
     * <ul>
     * <li>two shared libraries are installed, one version per SL.</li>
     * </ul>
     * using:
     * <ul>
     * <li>its name,</li>
     * <li>its version,</li>
     * <li>the type is not set.</li>
     * </ul>
     * </p>
     * <p>
     * Expected result: No error occurs, the SL is retrieved.
     * </p>
     */
    @Test
    public void getArtifactSLFromNameAndVersionIntoEnvWithTwoDifferentSls() throws InstallationServiceErrorException,
            NotRemoteJMXClientException, ContainerAdministrationException, DuplicatedServiceException,
            MissingServiceException, UnsupportedArtifactTypeException, ArtifactNotFoundException,
            ArtifactTypeIsNeededException, ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");
        final SharedLibrary sl1 = new SharedLibrary(SL1_NAME, SL1_VERSION);
        this.jmxApi.addSharedLibrary(sl1);
        final SharedLibrary sl2 = new SharedLibrary(SL2_NAME, SL2_VERSION);
        this.jmxApi.addSharedLibrary(sl2);

        final ArtifactAdministration artifactAdm = containerAdm.newArtifactAdministration();
        assertTrue(artifactAdm instanceof JMXArtifactAdministration);

        final Artifact expectedArtifact = artifactAdm.getArtifact(null, SL1_NAME, SL1_VERSION);

        assertNotNull(expectedArtifact);
        assertTrue(expectedArtifact instanceof org.ow2.petals.admin.api.artifact.SharedLibrary);
        assertEquals(SL1_NAME, expectedArtifact.getName());
        assertEquals(SL1_VERSION, expectedArtifact.getVersion());
        assertEquals(org.ow2.petals.admin.api.artifact.SharedLibrary.TYPE, expectedArtifact.getType());
    }

}
