/**
 * Copyright (c) 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.jmx;

import static org.easymock.EasyMock.anyBoolean;
import static org.easymock.EasyMock.eq;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.ow2.petals.jmx.api.mock.junit.PetalsJmxApiJunitRule.ComponentType.BINDING;
import static org.ow2.petals.jmx.api.mock.junit.PetalsJmxApiJunitRule.ComponentType.ENGINE;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.easymock.EasyMock;
import org.easymock.IAnswer;
import org.junit.Rule;
import org.junit.Test;
import org.ow2.petals.admin.api.ContainerAdministration;
import org.ow2.petals.admin.api.PetalsAdministration;
import org.ow2.petals.admin.api.PetalsAdministrationFactory;
import org.ow2.petals.admin.api.artifact.Component;
import org.ow2.petals.admin.api.artifact.Component.ComponentType;
import org.ow2.petals.admin.api.artifact.ServiceAssembly;
import org.ow2.petals.admin.api.artifact.SharedLibrary;
import org.ow2.petals.admin.api.exception.ContainerAdministrationException;
import org.ow2.petals.admin.api.exception.DuplicatedServiceException;
import org.ow2.petals.admin.api.exception.MissingServiceException;
import org.ow2.petals.admin.jmx.topology.TopologyUtils;
import org.ow2.petals.admin.topology.Container;
import org.ow2.petals.admin.topology.Container.PortType;
import org.ow2.petals.admin.topology.Container.State;
import org.ow2.petals.admin.topology.Domain;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.jmx.api.api.TopologyServiceClient;
import org.ow2.petals.jmx.api.mock.junit.PetalsJmxApiJunitRule;

/**
 * Unit tests of {@link JMXContainerAdministration}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class JMXContainerAdministrationTest {

    @Rule
    final public PetalsJmxApiJunitRule jmxApiContainer0 = new PetalsJmxApiJunitRule();

    @Rule
    final public PetalsJmxApiJunitRule jmxApiContainer1 = new PetalsJmxApiJunitRule();

    @Rule
    final public PetalsJmxApiJunitRule jmxApiContainer1Fail = new PetalsJmxApiJunitRule(false);

    @Rule
    final public PetalsJmxApiJunitRule jmxApiContainer2 = new PetalsJmxApiJunitRule();

    /**
     * Check the {@link JMXContainerAdministration#getTopology(String, String, boolean)} against container filtering
     * where:
     * <ul>
     * <li>the topology is composed of:
     * <ul>
     * <li>a domain 'my-domain' containing the containers 'container-0-0', 'container-0-1' and 'container-0-2',</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>when filtering with the regexp 'container-.*':
     * <ul>
     * <li>no error,</li>
     * <li>all containers are returned.</li>
     * </ul>
     * </li>
     * <li>when filtering with the regexp 'container-[01]-1':
     * <ul>
     * <li>no error,</li>
     * <li>only the containers 'container-0-1' is returned.</li>
     * </ul>
     * </li>
     * <li>when filtering with the regexp 'container-0-1':
     * <ul>
     * <li>no error,</li>
     * <li>only the container 'container-0-1' is returned.</li>
     * </ul>
     * </li>
     * </ul>
     * </p>
     */
    @Test
    public void getTopology() throws ContainerAdministrationException, DuplicatedServiceException,
            MissingServiceException, PetalsException {

        final String myDomainName = "myDomain";
        final String container0PassPhrase = "container-0-passphrase";
        final String container0Name = "container-0-0";
        final String container0SLName = "container-0-sl-name";
        final String container0SLVersion = "1.0";
        final String container0CompName = "container-0-component-name";
        final String container0SAName = "container-0-service-assembly-name";
        final String container1Name = "container-0-1";
        final String container1SLName = "container-1-sl-name";
        final String container1SLVersion = "1.0";
        final String container1Comp0Name = "container-1-component-0-name";
        final String container1Comp1Name = "container-1-component-1-name";
        final String container1SA0Name = "container-1-service-assembly-0-name";
        final String container1SA1Name = "container-1-service-assembly-1-name";
        final String container2Name = "container-0-2";
        final String container2Comp0Name = "container-2-component-0-name";

        final TopologyServiceClient jmxTopologyServiceClientMock0 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.expect(jmxTopologyServiceClientMock0.retrieveTopology(eq(container0PassPhrase), anyBoolean()))
                .andAnswer(new IAnswer<Set<Map<String, String>>>() {

                    @Override
                    public Set<Map<String, String>> answer() throws Throwable {
                        final Domain domain = new Domain(myDomainName);
                        final Map<PortType, Integer> container0Ports = new HashMap<>();
                        container0Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer0.getPort());
                        final Container container0 = new Container(container0Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer0.getHost().getHostName(),
                                container0Ports, JMXContainerAdministrationTest.this.jmxApiContainer0.getUsername(),
                                JMXContainerAdministrationTest.this.jmxApiContainer0.getPassword(), State.REACHABLE);
                        final Map<PortType, Integer> container1Ports = new HashMap<>();
                        container1Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer1.getPort());
                        final Container container1 = new Container(container1Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer1.getHost().getHostName(),
                                container1Ports, JMXContainerAdministrationTest.this.jmxApiContainer1.getUsername(),
                                JMXContainerAdministrationTest.this.jmxApiContainer1.getPassword(), State.REACHABLE);
                        final Map<PortType, Integer> container2Ports = new HashMap<>();
                        container2Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer2.getPort());
                        final Container container2 = new Container(container2Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer2.getHost().getHostName(),
                                container2Ports, JMXContainerAdministrationTest.this.jmxApiContainer2.getUsername(),
                                JMXContainerAdministrationTest.this.jmxApiContainer2.getPassword(), State.REACHABLE);
                        domain.addContainers(container0, container1, container2);
                        return TopologyUtils.createTopologyMapSet(domain);
                    }
                }).anyTimes();
        EasyMock.replay(jmxTopologyServiceClientMock0);
        this.jmxApiContainer0.setTopologyServiceClient(jmxTopologyServiceClientMock0);
        this.jmxApiContainer0.addSharedLibrary(
                new org.ow2.petals.jmx.api.mock.junit.SharedLibrary(container0SLName, container0SLVersion));
        this.jmxApiContainer0.addComponentClient(container0CompName, BINDING);
        this.jmxApiContainer0.addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container0SAName));

        final TopologyServiceClient jmxTopologyServiceClientMock1 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.replay(jmxTopologyServiceClientMock1);
        this.jmxApiContainer1.setTopologyServiceClient(jmxTopologyServiceClientMock1);
        org.ow2.petals.jmx.api.mock.junit.SharedLibrary container1SL = new org.ow2.petals.jmx.api.mock.junit.SharedLibrary(
                container1SLName, container1SLVersion);
        this.jmxApiContainer1.addSharedLibrary(container1SL);
        this.jmxApiContainer1.addComponentClient(container1Comp0Name, BINDING);
        this.jmxApiContainer1.addComponentClient(container1Comp1Name, ENGINE,
                new org.ow2.petals.jmx.api.mock.junit.SharedLibrary[] { container1SL });
        this.jmxApiContainer1.addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container1SA0Name));
        this.jmxApiContainer1.addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container1SA1Name));

        final TopologyServiceClient jmxTopologyServiceClientMock2 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.replay(jmxTopologyServiceClientMock2);
        this.jmxApiContainer2.setTopologyServiceClient(jmxTopologyServiceClientMock2);
        this.jmxApiContainer2.addComponentClient(container2Comp0Name, BINDING);

        final PetalsAdministration container0AdmApi = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        container0AdmApi.connect("localhost", this.jmxApiContainer0.getPort(), this.jmxApiContainer0.getUsername(),
                "petals");
        final ContainerAdministration container0Adm = container0AdmApi.newContainerAdministration();
        assertTrue(container0Adm instanceof JMXContainerAdministration);

        try {
            // Check against container name regexp: 'container-.*', without requesting JBI artifacts
            {
                final Domain domain = container0Adm.getTopology("container-.*", container0PassPhrase, false);
                assertEquals(myDomainName, domain.getName());

                {
                    final List<Container> containers = domain.getContainers();
                    assertEquals(3, containers.size());
                    Container container00 = null;
                    Container container01 = null;
                    Container container02 = null;
                    for (final Container container : containers) {
                        if (container0Name.equals(container.getContainerName())) {
                            container00 = container;
                        } else if (container1Name.equals(container.getContainerName())) {
                            container01 = container;
                        } else if (container2Name.equals(container.getContainerName())) {
                            container02 = container;
                        } else {
                            fail("Unexpected container found: " + container.getContainerName());
                        }
                    }
                    assertNotNull("Container #00 is missing", container00);
                    assertNotNull("Container #01 is missing", container01);
                    assertNotNull("Container #02 is missing", container02);
                }
            }

            // Check against container name regexp: 'container-.*', with requesting JBI artifacts
            {
                final Domain domain = container0Adm.getTopology("container-.*", container0PassPhrase, true);
                assertEquals(myDomainName, domain.getName());

                {
                    final List<Container> containers = domain.getContainers();
                    assertEquals(3, containers.size());
                    for (final Container container : containers) {
                        if (container0Name.equals(container.getContainerName())) {
                            final List<SharedLibrary> sharedLibraries = container.getSharedLibraries();
                            assertEquals(1, sharedLibraries.size());
                            assertEquals(container0SLName, sharedLibraries.get(0).getName());
                            assertEquals(container0SLVersion, sharedLibraries.get(0).getVersion());
                            final List<Component> components = container.getComponents();
                            assertEquals(1, components.size());
                            assertEquals(container0CompName, components.get(0).getName());
                            assertEquals(ComponentType.BC, components.get(0).getComponentType());
                            assertEquals(0, components.get(0).getSharedLibraries().size());
                            final List<ServiceAssembly> serviceAssemblies = container.getServiceAssemblies();
                            assertEquals(1, serviceAssemblies.size());
                            assertEquals(container0SAName, serviceAssemblies.get(0).getName());
                        } else if (container1Name.equals(container.getContainerName())) {
                            final List<SharedLibrary> sharedLibraries = container.getSharedLibraries();
                            assertEquals(1, sharedLibraries.size());
                            assertEquals(container1SLName, sharedLibraries.get(0).getName());
                            assertEquals(container1SLVersion, sharedLibraries.get(0).getVersion());
                            final List<Component> components = container.getComponents();
                            assertEquals(2, components.size());
                            assertEquals(container1Comp0Name, components.get(0).getName());
                            assertEquals(ComponentType.BC, components.get(0).getComponentType());
                            assertEquals(0, components.get(0).getSharedLibraries().size());
                            assertEquals(container1Comp1Name, components.get(1).getName());
                            assertEquals(ComponentType.SE, components.get(1).getComponentType());
                            assertEquals(1, components.get(1).getSharedLibraries().size());
                            assertEquals(sharedLibraries.get(0),
                                    components.get(1).getSharedLibraries().iterator().next());
                            final List<ServiceAssembly> serviceAssemblies = container.getServiceAssemblies();
                            assertEquals(2, serviceAssemblies.size());
                            assertEquals(container1SA0Name, serviceAssemblies.get(0).getName());
                            assertEquals(container1SA1Name, serviceAssemblies.get(1).getName());
                        } else if (container2Name.equals(container.getContainerName())) {
                            assertEquals(0, container.getSharedLibraries().size());
                            final List<Component> components = container.getComponents();
                            assertEquals(1, components.size());
                            assertEquals(container2Comp0Name, components.get(0).getName());
                            assertEquals(ComponentType.BC, components.get(0).getComponentType());
                            assertEquals(0, components.get(0).getSharedLibraries().size());
                            assertEquals(0, container.getServiceAssemblies().size());
                        } else {
                            fail("Unexpected container found: " + container.getContainerName());
                        }
                    }
                }
            }

            // Check against container name regexp: 'container-[01]-1', without requesting JBI artifacts
            {
                final Domain domain = container0Adm.getTopology("container-[01]-1", container0PassPhrase, false);
                assertEquals(myDomainName, domain.getName());

                {
                    final List<Container> containers = domain.getContainers();
                    assertEquals(1, containers.size());
                    assertEquals("Container #01 is missing", container1Name, containers.get(0).getContainerName());
                }
            }

            // Check against container name regexp: 'container-0-1', without requesting JBI artifacts
            {
                final Domain domain = container0Adm.getTopology("container-0-1", container0PassPhrase, false);
                assertEquals(myDomainName, domain.getName());
                final List<Container> containers = domain.getContainers();
                assertEquals(1, containers.size());
                assertEquals("Container #01 is missing", container1Name, containers.get(0).getContainerName());
            }
        } finally {
            container0AdmApi.disconnect();
        }

    }

    /**
     * Check the {@link JMXContainerAdministration#getTopology(String, String, boolean)} against getting artifacts
     * where:
     * <ul>
     * <li>the topology is composed of:
     * <ul>
     * <li>a domain 'my-domain' containing the containers 'container-0-0', 'container-0-1' and 'container-0-2',</li>
     * </ul>
     * </li>
     * <li>the Petals CLI command is executed on 'container-0-0',</li>
     * <li>a connection error occurs calling 'container-0-1',</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error occurs,</li>
     * <li>artifacts of 'container-0-0' are returned,</li>
     * <li>artifacts of 'container-0-1' are not returned,</li>
     * <li>artifacts of 'container-0-2' are returned.</li>
     * <ul>
     * </p>
     */
    @Test
    public void getTopologyWithAllArtifactsWithConnectionError() throws ContainerAdministrationException,
            DuplicatedServiceException, MissingServiceException, PetalsException {

        final String myDomainName = "myDomain";
        final String container0PassPhrase = "container-0-passphrase";
        final String container0Name = "container-0-0";
        final String container0SLName = "container-0-sl-name";
        final String container0SLVersion = "1.0";
        final String container0CompName = "container-0-component-name";
        final String container0SAName = "container-0-service-assembly-name";
        final String container1Name = "container-0-1";
        final String container1Comp0Name = "container-1-component-0-name";
        final String container1Comp1Name = "container-1-component-1-name";
        final String container1SA0Name = "container-1-service-assembly-0-name";
        final String container1SA1Name = "container-1-service-assembly-1-name";
        final String container2Name = "container-0-2";
        final String container2Comp0Name = "container-2-component-0-name";

        final TopologyServiceClient jmxTopologyServiceClientMock0 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.expect(jmxTopologyServiceClientMock0.retrieveTopology(container0PassPhrase, true))
                .andAnswer(new IAnswer<Set<Map<String, String>>>() {

                    @Override
                    public Set<Map<String, String>> answer() throws Throwable {
                        final Domain domain = new Domain(myDomainName);
                        final Map<PortType, Integer> container0Ports = new HashMap<>();
                        container0Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer0.getPort());
                        final Container container0 = new Container(container0Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer0.getHost().getHostName(),
                                container0Ports, JMXContainerAdministrationTest.this.jmxApiContainer0.getUsername(),
                                JMXContainerAdministrationTest.this.jmxApiContainer0.getPassword(), State.REACHABLE);
                        final Map<PortType, Integer> container1Ports = new HashMap<>();
                        container1Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer1Fail.getPort());
                        final Container container1 = new Container(container1Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer1Fail.getHost().getHostName(),
                                container1Ports, JMXContainerAdministrationTest.this.jmxApiContainer1Fail.getUsername(),
                                JMXContainerAdministrationTest.this.jmxApiContainer1Fail.getPassword(),
                                State.UNREACHABLE);
                        final Map<PortType, Integer> container2Ports = new HashMap<>();
                        container2Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer1.getPort());
                        final Container container2 = new Container(container2Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer1.getHost().getHostName(),
                                container2Ports, JMXContainerAdministrationTest.this.jmxApiContainer1.getUsername(),
                                JMXContainerAdministrationTest.this.jmxApiContainer1.getPassword(), State.REACHABLE);
                        domain.addContainers(container0, container1, container2);
                        return TopologyUtils.createTopologyMapSet(domain);
                    }
                }).anyTimes();
        EasyMock.replay(jmxTopologyServiceClientMock0);
        this.jmxApiContainer0.setTopologyServiceClient(jmxTopologyServiceClientMock0);
        this.jmxApiContainer0.addSharedLibrary(
                new org.ow2.petals.jmx.api.mock.junit.SharedLibrary(container0SLName, container0SLVersion));
        this.jmxApiContainer0.addComponentClient(container0CompName, BINDING);
        this.jmxApiContainer0
                .addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container0SAName));

        final TopologyServiceClient jmxTopologyServiceClientMock1 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.replay(jmxTopologyServiceClientMock1);
        this.jmxApiContainer1Fail.setTopologyServiceClient(jmxTopologyServiceClientMock1);
        this.jmxApiContainer1Fail.addComponentClient(container1Comp0Name, BINDING);
        this.jmxApiContainer1Fail.addComponentClient(container1Comp1Name, ENGINE);
        this.jmxApiContainer1Fail
                .addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container1SA0Name));
        this.jmxApiContainer1Fail
                .addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container1SA1Name));

        final TopologyServiceClient jmxTopologyServiceClientMock2 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.replay(jmxTopologyServiceClientMock2);
        this.jmxApiContainer1.setTopologyServiceClient(jmxTopologyServiceClientMock2);
        this.jmxApiContainer1.addComponentClient(container2Comp0Name, BINDING);

        // Execute test
        PetalsAdministration container0AdmApi = PetalsAdministrationFactory.getInstance().newPetalsAdministrationAPI();
        container0AdmApi.connect("localhost", this.jmxApiContainer0.getPort(), this.jmxApiContainer0.getUsername(),
                "petals");

        final ContainerAdministration container0Adm = container0AdmApi.newContainerAdministration();
        assertTrue(container0Adm instanceof JMXContainerAdministration);

        final Domain domain = container0Adm.getTopology(null, container0PassPhrase, true);

        // Result assertions
        assertTrue(container0AdmApi.isConnected());
        container0AdmApi.disconnect();

        assertEquals(myDomainName, domain.getName());

        final List<Container> containers = domain.getContainers();
        assertEquals(3, containers.size());
        for (final Container container : containers) {
            if (container0Name.equals(container.getContainerName())) {
                final List<SharedLibrary> sharedLibraries = container.getSharedLibraries();
                assertEquals(1, sharedLibraries.size());
                assertEquals(container0SLName, sharedLibraries.get(0).getName());
                assertEquals(container0SLVersion, sharedLibraries.get(0).getVersion());
                final List<Component> components = container.getComponents();
                assertEquals(1, components.size());
                assertEquals(container0CompName, components.get(0).getName());
                assertEquals(ComponentType.BC, components.get(0).getComponentType());
                final List<ServiceAssembly> serviceAssemblies = container.getServiceAssemblies();
                assertEquals(1, serviceAssemblies.size());
                assertEquals(container0SAName, serviceAssemblies.get(0).getName());
            } else if (container1Name.equals(container.getContainerName())) {
                // No Jbi artifacts returned because the container is unreachable
                assertEquals(0, container.getSharedLibraries().size());
                assertEquals(0, container.getComponents().size());
                assertEquals(0, container.getServiceAssemblies().size());
            } else if (container2Name.equals(container.getContainerName())) {
                assertEquals(0, container.getSharedLibraries().size());
                final List<Component> components = container.getComponents();
                assertEquals(1, components.size());
                assertEquals(container2Comp0Name, components.get(0).getName());
                assertEquals(ComponentType.BC, components.get(0).getComponentType());
                assertEquals(0, container.getServiceAssemblies().size());
            } else {
                fail("Unexpected container found: " + container.getContainerName());
            }
        }

    }

    /**
     * Check the {@link JMXContainerAdministration#getTopology(String, String, boolean)} against getting artifacts
     * where:
     * <ul>
     * <li>the topology is composed of:
     * <ul>
     * <li>a domain 'my-domain' containing the containers 'container-0-0', 'container-0-1' and 'container-0-2',</li>
     * </ul>
     * </li>
     * <li>the Petals CLI command is executed on 'container-0-0',</li>
     * <li>the passphrase is invalid (ie. JMX credentials are not returned by
     * {@link TopologyServiceClient#retrieveTopology(String, boolean)},</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>an error occurs</li>
     * <ul>
     * </p>
     */
    @Test
    public void getTopologyWithAllArtifactsInvalidPassphrase() throws ContainerAdministrationException,
            DuplicatedServiceException, MissingServiceException, PetalsException {

        final String myDomainName = "myDomain";
        final String container0PassPhrase = "container-0-passphrase";
        final String container0Name = "container-0-0";
        final String container0SLName = "container-0-sl-name";
        final String container0SLVersion = "1.0";
        final String container0CompName = "container-0-component-name";
        final String container0SAName = "container-0-service-assembly-name";
        final String container1Name = "container-0-1";
        final String container1Comp0Name = "container-1-component-0-name";
        final String container1Comp1Name = "container-1-component-1-name";
        final String container1SA0Name = "container-1-service-assembly-0-name";
        final String container1SA1Name = "container-1-service-assembly-1-name";
        final String container2Name = "container-0-2";
        final String container2Comp0Name = "container-2-component-0-name";

        final TopologyServiceClient jmxTopologyServiceClientMock0 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.expect(jmxTopologyServiceClientMock0.retrieveTopology(container0PassPhrase, true))
                .andAnswer(new IAnswer<Set<Map<String, String>>>() {

                    @Override
                    public Set<Map<String, String>> answer() throws Throwable {
                        final Domain domain = new Domain(myDomainName);
                        final Map<PortType, Integer> container0Ports = new HashMap<>();
                        container0Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer0.getPort());
                        final Container container0 = new Container(container0Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer0.getHost().getHostName(),
                                container0Ports, null, null, State.REACHABLE);
                        final Map<PortType, Integer> container1Ports = new HashMap<>();
                        container1Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer1.getPort());
                        final Container container1 = new Container(container1Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer1.getHost().getHostName(),
                                container1Ports, null, null, State.REACHABLE);
                        final Map<PortType, Integer> container2Ports = new HashMap<>();
                        container2Ports.put(PortType.JMX,
                                JMXContainerAdministrationTest.this.jmxApiContainer2.getPort());
                        final Container container2 = new Container(container2Name,
                                JMXContainerAdministrationTest.this.jmxApiContainer2.getHost().getHostName(),
                                container2Ports, null, null, State.REACHABLE);
                        domain.addContainers(container0, container1, container2);
                        return TopologyUtils.createTopologyMapSet(domain);
                    }
                }).anyTimes();
        EasyMock.replay(jmxTopologyServiceClientMock0);
        this.jmxApiContainer0.setTopologyServiceClient(jmxTopologyServiceClientMock0);
        this.jmxApiContainer0.addSharedLibrary(
                new org.ow2.petals.jmx.api.mock.junit.SharedLibrary(container0SLName, container0SLVersion));
        this.jmxApiContainer0.addComponentClient(container0CompName, BINDING);
        this.jmxApiContainer0
                .addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container0SAName));

        final TopologyServiceClient jmxTopologyServiceClientMock1 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.replay(jmxTopologyServiceClientMock1);
        this.jmxApiContainer1.setTopologyServiceClient(jmxTopologyServiceClientMock1);
        this.jmxApiContainer1.addComponentClient(container1Comp0Name, BINDING);
        this.jmxApiContainer1.addComponentClient(container1Comp1Name, ENGINE);
        this.jmxApiContainer1
                .addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container1SA0Name));
        this.jmxApiContainer1
                .addServiceAssembly(new org.ow2.petals.jmx.api.mock.junit.ServiceAssembly(container1SA1Name));

        final TopologyServiceClient jmxTopologyServiceClientMock2 = EasyMock.createMock(TopologyServiceClient.class);
        EasyMock.replay(jmxTopologyServiceClientMock2);
        this.jmxApiContainer2.setTopologyServiceClient(jmxTopologyServiceClientMock2);
        this.jmxApiContainer2.addComponentClient(container2Comp0Name, BINDING);

        // Execute test
        PetalsAdministration container0AdmApi = PetalsAdministrationFactory.getInstance().newPetalsAdministrationAPI();
        container0AdmApi.connect("localhost", this.jmxApiContainer0.getPort(), this.jmxApiContainer0.getUsername(),
                "petals");
        final ContainerAdministration container0Adm = container0AdmApi.newContainerAdministration();
        assertTrue(container0Adm instanceof JMXContainerAdministration);

        container0Adm.getTopology("container-.*", container0PassPhrase, true);
    }

}
