/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.jmx.artifact;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;

import javax.management.NotCompliantMBeanException;
import javax.management.StandardMBean;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.admin.api.PetalsAdministration;
import org.ow2.petals.admin.api.PetalsAdministrationFactory;
import org.ow2.petals.admin.api.artifact.Component;
import org.ow2.petals.admin.api.artifact.Component.ComponentType;
import org.ow2.petals.admin.api.artifact.lifecycle.ComponentLifecycle;
import org.ow2.petals.admin.api.exception.ArtifactAdministrationException;
import org.ow2.petals.admin.api.exception.ArtifactDeployedException;
import org.ow2.petals.admin.api.exception.ArtifactNotFoundException;
import org.ow2.petals.admin.api.exception.ArtifactUnsupportedOperationException;
import org.ow2.petals.admin.api.exception.ContainerAdministrationException;
import org.ow2.petals.admin.api.exception.DuplicatedServiceException;
import org.ow2.petals.admin.api.exception.MissingServiceException;
import org.ow2.petals.jmx.api.api.ComponentClient;
import org.ow2.petals.jmx.api.api.exception.ComponentDoesNotExistException;
import org.ow2.petals.jmx.api.api.exception.ComponentErrorException;
import org.ow2.petals.jmx.api.api.exception.ConnectionErrorException;
import org.ow2.petals.jmx.api.api.exception.InstallationServiceErrorException;
import org.ow2.petals.jmx.api.api.exception.NotRemoteJMXClientException;
import org.ow2.petals.jmx.api.mock.junit.PetalsJmxApiJunitRule;
import org.ow2.petals.jmx.api.mock.junit.configuration.component.InstallationConfigurationServiceClientMock;

/**
 * Unit tests of {@link JMXComponentLifecycle}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class JMXComponentLifecycleTest {

    private static final String COMPONENT_ID = "my-component-identifier";

    @Rule
    public final TemporaryFolder temporaryFolder = new TemporaryFolder();

    @Rule
    public final PetalsJmxApiJunitRule jmxApi = new PetalsJmxApiJunitRule();

    /**
     * Check that a component having no installer configuration MBean can be deployed
     */
    @Test
    public void deployComponentWithoutInstallerConfigurationMBean() throws Exception {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");

        final File componentArchive = this.temporaryFolder.newFile(COMPONENT_ID + ".zip");
        final Component component = new Component(COMPONENT_ID, ComponentType.BC);
        final ComponentLifecycle componentLC = containerAdm.newArtifactLifecycleFactory()
                .createComponentLifecycle(component);
        assertTrue(componentLC instanceof JMXComponentLifecycle);
        componentLC.deploy(componentArchive.toURI().toURL());
    }

    @Test
    public void setInstallParameters() throws Exception {
        ComponentConfig config = new ComponentConfig();
        this.jmxApi.addComponentInstallerClient(COMPONENT_ID,
                org.ow2.petals.jmx.api.mock.junit.PetalsJmxApiJunitRule.ComponentType.BINDING,
                new InstallationConfigurationServiceClientMock(config), null);

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");

        final Component component = new Component(COMPONENT_ID, ComponentType.BC);
        component.getParameters().put("Param", "10");

        final ComponentLifecycle componentLC = containerAdm.newArtifactLifecycleFactory()
                .createComponentLifecycle(component);
        assertTrue(componentLC instanceof JMXComponentLifecycle);
        componentLC.setParameters();

        assertEquals(config.getParam(), 10);
    }

    public interface InstallerConfig {
        public int getParam();

        public void setParam(int param);
    }

    public static class ComponentConfig extends StandardMBean implements InstallerConfig {
        private int param = 5;

        public ComponentConfig() throws NotCompliantMBeanException {
            super(InstallerConfig.class);
        }

        @Override
        public int getParam() {
            return param;
        }

        @Override
        public void setParam(int param) {
            this.param = param;
        }
    }

    /**
     * Check that an error occurs when deploying a component having no installer configuration MBean and trying to set
     * parameter.
     */
    @Test(expected = ArtifactAdministrationException.class)
    public void deployComponentWithoutInstallerConfigurationMBeanWithParam()
            throws DuplicatedServiceException, MissingServiceException, NotRemoteJMXClientException,
            ContainerAdministrationException, IOException, ArtifactDeployedException, ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");

        final File componentArchive = this.temporaryFolder.newFile(COMPONENT_ID + ".zip");
        final Component component = new Component(COMPONENT_ID, ComponentType.BC);
        component.getParameters().setProperty("param-name", "param-value");
        final ComponentLifecycle componentLC = containerAdm.newArtifactLifecycleFactory()
                .createComponentLifecycle(component);
        assertTrue(componentLC instanceof JMXComponentLifecycle);
        componentLC.deploy(componentArchive.toURI().toURL());
    }

    /**
     * Check the placeholder configuration reloading when the given component is started.
     */
    @Test
    public void reloadConfigurationComponentStarted() throws DuplicatedServiceException, MissingServiceException,
            NotRemoteJMXClientException, ContainerAdministrationException, ArtifactNotFoundException,
            ArtifactUnsupportedOperationException, ArtifactAdministrationException, InstallationServiceErrorException,
            ComponentDoesNotExistException, ComponentErrorException, ConnectionErrorException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");

        this.jmxApi.addComponentClient(COMPONENT_ID, PetalsJmxApiJunitRule.ComponentType.BINDING);
        final ComponentClient componentService = this.jmxApi.getComponentClient(COMPONENT_ID);
        assertEquals(componentService.getCurrentState(), ComponentClient.State.STARTED);

        final Component component = new Component(COMPONENT_ID, ComponentType.BC);
        final ComponentLifecycle componentLC = containerAdm.newArtifactLifecycleFactory()
                .createComponentLifecycle(component);
        assertTrue(componentLC instanceof JMXComponentLifecycle);
        componentLC.reloadConfiguration();
    }

    /**
     * Check the placeholder configuration reloading when the given component is stopped.
     * 
     * @throws InstallationServiceErrorException
     */
    @Test
    public void reloadConfigurationComponentStopped() throws DuplicatedServiceException, MissingServiceException,
            NotRemoteJMXClientException, ContainerAdministrationException, ArtifactNotFoundException,
            ArtifactUnsupportedOperationException, ArtifactAdministrationException, InstallationServiceErrorException,
            ComponentDoesNotExistException, ComponentErrorException, ConnectionErrorException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");

        this.jmxApi.addComponentClient(COMPONENT_ID, PetalsJmxApiJunitRule.ComponentType.BINDING);
        final ComponentClient componentService = this.jmxApi.getComponentClient(COMPONENT_ID);
        componentService.stop();
        assertEquals(componentService.getCurrentState(), ComponentClient.State.STOPPED);

        final Component component = new Component(COMPONENT_ID, ComponentType.BC);
        final ComponentLifecycle componentLC = containerAdm.newArtifactLifecycleFactory()
                .createComponentLifecycle(component);
        assertTrue(componentLC instanceof JMXComponentLifecycle);
        componentLC.reloadConfiguration();
    }

    /**
     * Check the placeholder configuration reloading when the given component is not found.
     */
    @Test(expected = ArtifactNotFoundException.class)
    public void reloadConfigurationComponentNotFound() throws DuplicatedServiceException, MissingServiceException,
            NotRemoteJMXClientException, ContainerAdministrationException, ArtifactNotFoundException,
            ArtifactUnsupportedOperationException, ArtifactAdministrationException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");

        final Component component = new Component("component-not-found", ComponentType.BC);
        final ComponentLifecycle componentLC = containerAdm.newArtifactLifecycleFactory()
                .createComponentLifecycle(component);
        assertTrue(componentLC instanceof JMXComponentLifecycle);
        componentLC.reloadConfiguration();
    }

    /**
     * Check the placeholder configuration reloading when the given component is started but does not support the
     * configuration reloading.
     */
    @Test(expected = ArtifactUnsupportedOperationException.class)
    public void reloadConfigurationComponentDoesNotSupportReload()
            throws DuplicatedServiceException, MissingServiceException, NotRemoteJMXClientException,
            ContainerAdministrationException, ArtifactNotFoundException, ArtifactUnsupportedOperationException,
            ArtifactAdministrationException, InstallationServiceErrorException, ComponentDoesNotExistException,
            ComponentErrorException, ConnectionErrorException {

        final PetalsAdministration containerAdm = PetalsAdministrationFactory.getInstance()
                .newPetalsAdministrationAPI();
        containerAdm.connect("localhost", this.jmxApi.getPort(), this.jmxApi.getUsername(), "petals");

        this.jmxApi.addComponentClient(COMPONENT_ID, PetalsJmxApiJunitRule.ComponentType.BINDING)
                .withUnsupportedReloadConfiguration();
        final ComponentClient componentService = this.jmxApi.getComponentClient(COMPONENT_ID);
        assertEquals(componentService.getCurrentState(), ComponentClient.State.STARTED);

        final Component component = new Component(COMPONENT_ID, ComponentType.BC);
        final ComponentLifecycle componentLC = containerAdm.newArtifactLifecycleFactory()
                .createComponentLifecycle(component);
        assertTrue(componentLC instanceof JMXComponentLifecycle);
        componentLC.reloadConfiguration();
    }
}
